# This PCell instantiates two ebeam_dc_halfring_straight

import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class ring_doublebus(pya.PCellDeclarationHelper):
  """
  The PCell uses two ebeam_dc_halfring_straight PCells.
  """

  def __init__(self):
    # Important: initialize the super class
    super(ring_doublebus, self).__init__()
    self.technology_name = 'EBeam'
    TECHNOLOGY = get_technology_by_name(self.technology_name)
    
    
    # declare the parameters
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("r", self.TypeDouble, "Radius", default = 10)
    self.param("w", self.TypeDouble, "Waveguide Width", default = 0.5)
    self.param("g", self.TypeDouble, "Gap", default = 0.2)
    self.param("Lc", self.TypeDouble, "Coupler Length", default = 0.0)
    self.param("orthogonal_identifier", self.TypeInt, "Orthogonal identifier (1=TE, 2=TM)", default = 1)     
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = TECHNOLOGY['Text'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "ring_doublebus(R=" + ('%.3f' % self.r) + ",g=" + ('%g' % (1000*self.g)) + ",Lc=" + ('%g' % (1000*self.Lc)) + ",orthogonal_identifier=" + ('%s' % self.orthogonal_identifier) + ")"

  def can_create_from_shape_impl(self):
    return False
    
  def produce_impl(self):
    # This is the main part of the implementation: create the layout

    from SiEPIC.extend import to_itype

    # print(self.layout.technology_name)  # the technology is not assigned in a PCell, for some reason    
    self.layout.technology_name=self.technology_name
    
    pcell = self.layout.create_cell("ebeam_dc_halfring_straight", self.technology_name, {
                                                             "silayer": self.silayer,
                                                             "r": self.r,
                                                             "w": self.w,
                                                             "g": self.g,
                                                             "Lc": self.Lc,
                                                             "orthogonal_identifier": self.orthogonal_identifier,
                                                             "pinrec": self.pinrec,
                                                             "devrec": self.devrec,
                                                             "textl": self.textl,
                                                                })
    
    if pcell==None:
        raise Exception("problem! cannot create PCell from library: %s" % self.technology_name)
    inst = self.cell.insert(pya.CellInstArray(pcell.cell_index(), pya.Trans(pya.Trans.R0, 0, 0)))
    inst = self.cell.insert(pya.CellInstArray(pcell.cell_index(), pya.Trans(pya.Trans.R180, 0, to_itype(self.r*2+self.g*2+self.w*2,self.layout.dbu))))
    # print((self.r*2+self.g*2+self.w*2)/self.layout.dbu)
    # print("Done drawing the layout for - ring_doublebus: %.3f-%g" % ( self.r, self.g) )



import pya
from pya import *
from SiEPIC.utils import get_technology, get_technology_by_name
from SiEPIC.utils import arc, arc_wg, arc_to_waveguide, points_per_circle#,layout
import math


class phc_gc_hex(pya.PCellDeclarationHelper):
  """
  Input: length, width
  """
  import numpy
  
  def __init__(self):
    
    # Important: initialize the super class
    super(phc_gc_hex, self).__init__()

    self.param("a", self.TypeDouble, "lattice constant (microns)", default = 0.243)     
    self.param("x", self.TypeInt, "Number of holes in x direction", default = 78)    
    self.param("y", self.TypeInt, "Number of holes in y direction", default = 50)   
    self.param("r", self.TypeDouble, "hole radius (microns)", default = 0.0735)
    self.param("vertices",  self.TypeInt, "Number of vertices in circle", default = 32)
    TECHNOLOGY = get_technology_by_name('EBeam')
    self.param("positive",  self.TypeInt, "Positive", default = False)
    self.param("apodized",  self.TypeInt, "apodized", default = False)
    self.param("feature_size", self.TypeDouble, "minimum feature size (microns)", default = 0.06)  
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = TECHNOLOGY['Text'])
    self.param("invert", self.TypeLayer, "Layer to invert", default =TECHNOLOGY['Si'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "phc_gc_hex_a%s-r%.3f" % \
    (self.a, self.r)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    debug = False
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    
    LayerSi = self.layer
    LayerSiN = ly.layer(self.layer)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    LayerTextN = ly.layer(self.textl)
    LayerInvert=ly.layer(self.invert)
    n_vertices = int(self.vertices)
    a = self.a/dbu
    r = self.r/dbu
    n_x = int(math.ceil(self.x/2))
    n_y = int(math.ceil(self.y/2))
    positive=bool(self.positive)
    minimum_feature=self.feature_size
    apodized=bool(self.apodized)
    length_slab_x = 2*n_x*a
    
    length_slab_y = 2*(n_y-2)*a

    k = 2
    
    #function to creat polygon pts for right half of a hole in a hexagon unit cell
    
    def circle(x,y,r):
      npts = n_vertices
      theta = 2 * math.pi / npts # increment, in radians
      pts = []
      for i in range(0, npts):
        pts.append(Point.from_dpoint(pya.DPoint((x+r*math.cos(i*theta))/1, (y+r*math.sin(i*theta))/1)))
      return pts
      
    def hexagon_hole_half(a,r): 
      npts = 10    
      theta_div = math.pi/3
      theta_div_hole = math.pi/npts
      triangle_length = a/math.sqrt(3)
      pts = []
      for i in range(0,4):
        pts.append(Point.from_dpoint(pya.DPoint(triangle_length*math.cos(i*theta_div-math.pi/2), triangle_length*math.sin(i*theta_div-math.pi/2))))
      for i in range(0, npts+1):
        pts.append(Point.from_dpoint(pya.DPoint(r*math.cos(math.pi/2-i*theta_div_hole), r*math.sin(math.pi/2-i*theta_div_hole))))
      return pts
    
    def hexagon_shifthole_half(a,r): 
      npts = 10    
      theta_div = math.pi/3
      theta_div_hole = math.pi/npts
      triangle_length = a*1.235/math.sqrt(3)
      pts = []
      for i in range(0,4):
        pts.append(Point.from_dpoint(pya.DPoint(triangle_length*math.cos(i*theta_div-math.pi/2), triangle_length*math.sin(i*theta_div-math.pi/2))))
      for i in range(0, npts+1):
        pts.append(Point.from_dpoint(pya.DPoint(r*math.cos(math.pi/2-i*theta_div_hole), r*math.sin(math.pi/2-i*theta_div_hole))))
      return pts
    
    #function to creat polygon pts for right half of a hexagon unit cell      
    def hexagon_half(a): 
      theta_div = math.pi/3
      triangle_length = a/math.sqrt(3)
      pts = []
      for i in range(0,4):
        pts.append(Point.from_dpoint(pya.DPoint(triangle_length*math.cos(i*theta_div-math.pi/2), triangle_length*math.sin(i*theta_div-math.pi/2))))
      return pts      
    
    #create the right and left half of the hole and hexagon cells
    #hole_cell = pya.Region()
    #hexagon_cell = pya.Region()
    
        # Define Si slab and hole region for future subtraction
    Si_slab = pya.Region()
    Si_slab.insert(pya.Box(-length_slab_x/2+a*2, -length_slab_y/2, length_slab_x/2-a, length_slab_y/2))
    hole = pya.Region()
    hole_r = r
    
        # function to generate points to create a circle
    def circle(x,y,r):
      npts = n_vertices
      theta = 2 * math.pi / npts # increment, in radians
      pts = []
      for i in range(0, npts):
        pts.append(Point.from_dpoint(pya.DPoint((x+r*math.cos(i*theta))/1, (y+r*math.sin(i*theta))/1)))
      return pts
    import numpy  as np 
    def gaussian(x, mu, sig):
      return 1./(np.sqrt(2.*np.pi)*sig)*np.exp(-np.power((x - mu)/sig, 2.)/2) 
      
    # raster through all holes with shifts and waveguide 

    for j in range(-n_y,n_y+1):
      if j%2 == 0:
        skip=0
        apodization=0
        for i in range(-n_x,n_x+1):
          if i==0:
            continue
          if skip==0:
            skip=1
            continue
          elif skip==1:
            skip=2
            continue
          elif skip==2:
            skip=3 
            apodization=apodization+1
            continue
          elif skip==3:
            skip=1
            
            
          #radius=r/gaussian(float(i)/(2*(n_x+1)),2*(n_x+1),10)
          location=float(i)
          location=abs(location)
          #radius=r/gaussian(location/(2*(n_x+1)),2*(n_x+1),0.02)
          #radius=(((2*n_x)-abs(i+n_x+3)*r)/(2*n_x))
          radius=(float(apodization)/((n_x*2/3)-1))*r
          if radius<minimum_feature*500:
            radius=minimum_feature*500
          if apodized==False:
            radius=r
         
            #radius=minimum_feature
          hole_cell = circle(0,0,radius)
          hole_poly = pya.Polygon(hole_cell) 
          if debug: 
              print("x1 "+str(apodization))
          hole_x = abs(i)/i*(abs(i)-0.5)*a
          hole_y = j*a*math.sqrt(3)/2
          hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
          hole_t = hole_poly.transformed(hole_trans)
          hole.insert(hole_t) 
      elif j%2 == 1:
        skipodd=0
        apodization=0
        for i in range(-n_x,n_x+1):
          if i==-n_x:
            continue
          if i==n_x:
            continue
          if skipodd==0:
            skipodd=1
            continue
          elif skipodd==1:
            skipodd=2
            apodization=apodization+1
            continue
          elif skipodd==2:
            skipodd=3
          elif skipodd==3:
            skipodd=1
                        
          #radius=(((2*n_x)-abs(i+n_x+3)*r)/(2*n_x))
          radius=(float(apodization)/((n_x*2/3)-1))*r
          if radius<minimum_feature*500:
            radius=minimum_feature*500
          if apodized==False:
            radius=r
          
            #radius=minimum_feature
          #radius=r*(float(abs(i))/(2*(n_x+1)))
          hole_cell = circle(0,0,radius)
          hole_poly = pya.Polygon(hole_cell) 
          if debug:
              print("x2 "+str(apodization))
              print("p "+str(n_x))
          hole_x = i*a
          hole_y = j*a*math.sqrt(3)/2
          hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
          hole_t = hole_poly.transformed(hole_trans)
          hole.insert(hole_t) 
    if positive==True:        
        phc = hole
    else:
        phc = Si_slab - hole
    self.cell.shapes(LayerSiN).insert(phc)
    
    
     #print(hole_t_0)
    box_l = a/2
         
    # Pins on the waveguide:    
    pin_length = 200
    pin_w = a
    
    t = pya.Trans(Trans.R0, -length_slab_x/2+a*2,0)
    pin = pya.Path([pya.Point(-pin_length/2, 0), pya.Point(pin_length/2, 0)], pin_w)
    pin_t = pin.transformed(t)
    self.cell.shapes(LayerPinRecN).insert(pin_t)
    text = pya.Text ("pin1", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    
    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    points = [[-length_slab_x/2+a*2,0], [length_slab_x/2-a, 0]]
    points = [Point(each[0], each[1]) for each in points]
    path = Path(points, length_slab_y)   
    self.cell.shapes(LayerDevRecN).insert(path.simple_polygon())
    
    if positive==True: 
        self.cell.shapes(LayerInvert).insert(path.simple_polygon())
    
    return
    
    hole = pya.Region()
    
    hole_cell_pts = hexagon_hole_half(a,r)
    hexagon_pts = hexagon_half(a)
    hole_shiftcell_pts = hexagon_shifthole_half(a,r)
    hole_cell_poly_0 = pya.Polygon(hole_cell_pts)
    hexagon_cell_poly_0 = pya.Polygon(hexagon_pts)
    hole_shiftcell_poly_0 = pya.Polygon(hole_shiftcell_pts)

    hole_trans = pya.Trans(pya.Trans.R180)
    hole_cell_poly_1 = hole_cell_poly_0.transformed(hole_trans)
    hexagon_cell_poly_1 = hexagon_cell_poly_0.transformed(hole_trans)    
    hole_shiftcell_poly_1 = hole_shiftcell_poly_0.transformed(hole_trans)
    skip=1
    skip2=0   
    #create the photonic crystal with shifts and waveguides
    for j in range(-y+1,y):
      if j%2 == 0:
        for i in range(-x,x+1):
          
          #waveguide     
          if (j == k and i > 3):
            hole_x = abs(i)/i*(abs(i)-0.5)*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
            hole_t_1 = hole_cell_poly_1.transformed(hole_trans)            
            hole.insert(hole_t_0)
            hole.insert(hole_t_1)
          #filling the edges with half cell
                
                      
          elif i!=0:
            hole_x = abs(i)/i*(abs(i)-0.5)*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
            hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t_0)  
            hole.insert(hole_t_1)
      elif j%2 == 1:
        for i in range(-x,x+1):
          if skip==0 and i%3==0:
            skip=1
            continue
          if skip==1:
            skip=2
            continue 
          if skip==2:
            skip=0
            
          if i== -x:
            hole_x = abs(i)/i*(abs(i)-0.5)*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_1 = hexagon_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t_1)
          if i== x:
            hole_x = abs(i)/i*(abs(i)-0.5)*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hexagon_cell_poly_0.transformed(hole_trans)
            hole.insert(hole_t_0)
          #waveguide
          if (j == k and i > 3):
            hole_x =i*a
            hole_y =j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hexagon_cell_poly_0.transformed(hole_trans)
            hole_t_1 = hexagon_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t_0)
            hole.insert(hole_t_1)
          #filling the edges with half cell
          elif i in (-x,x):
            hole_x =i*a
            hole_y =j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y) 
            if i == -x:           
              hole_t = hole_cell_poly_0.transformed(hole_trans)
            else:
              hole_t = hole_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t)               
          
          else:  
            hole_x = i*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
            hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t_0)  
            hole.insert(hole_t_1)
    
    #print(hole_t_0)
    box_l = a/2
    
    cover_box = pya.Box(-length_slab_x/2, -a/2, length_slab_x/2, a/2) 
    box_y = y*a*math.sqrt(3)/2  
    cover_box_trans_0 = pya.Trans(Trans.R0, 0,box_y)
    cover_box_trans_1 = pya.Trans(Trans.R0, 0,-box_y)
    cover_box_t_0 = cover_box.transformed(cover_box_trans_0) 
    cover_box_t_1 = cover_box.transformed(cover_box_trans_1)     
    #hole.insert(pya.Box())    
    self.cell.shapes(LayerSiN).insert(hole)
    self.cell.shapes(LayerSiN).insert(cover_box_t_0)
    self.cell.shapes(LayerSiN).insert(cover_box_t_1)      
    # Pins on the waveguide:    
    pin_length = 200
    pin_w = a
    
    t = pya.Trans(Trans.R0, -length_slab_x/2,0)
    pin = pya.Path([pya.Point(-pin_length/2, 0), pya.Point(pin_length/2, 0)], pin_w)
    pin_t = pin.transformed(t)
    self.cell.shapes(LayerPinRecN).insert(pin_t)
    text = pya.Text ("pin1", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    points = [[-length_slab_x/2,0], [length_slab_x/2, 0]]
    points = [Point(each[0], each[1]) for each in points]
    path = Path(points, length_slab_y)   
    self.cell.shapes(LayerDevRecN).insert(path.simple_polygon())
      


import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class SWG_bentWG(pya.PCellDeclarationHelper):

    def __init__(self):

      # Important: initialize the super class
      super(SWG_bentWG, self).__init__()
      TECHNOLOGY = get_technology_by_name('EBeam')

      # declare the parameters
      self.param("pitch", self.TypeDouble, "Period [um]", default = 0.2)
      self.param("w", self.TypeDouble, "Waveguide Width [um]", default = 0.5)
      self.param("r", self.TypeDouble, "Radius [um]", default = 5) # minimum radius is 5 um in TE
      self.param("ff", self.TypeDouble, "Duty Cycle", default = 0.5)
      self.param("angle", self.TypeDouble, "Angle", default = 360)
      self.param("gap", self.TypeDouble, "Bus Gap [um]", default = 0.06)
      self.param("W_ratio", self.TypeDouble, "Lin/Lout", default = 1)
      self.param("doublebus", self.TypeBoolean, "Bus Type (Single=False, Double=True)", default = False)
      self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
      self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
      self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
  #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

    def display_text_impl(self):
      # Provide a descriptive text for the cell
      return "SWG_bentWG_%.3f" % \
      (self.pitch)

    def coerce_parameters_impl(self):
      pass

    def can_create_from_shape(self, layout, shape, layer):
      return False

    def produce_impl(self):
      debug = False
      
      from SiEPIC._globals import PIN_LENGTH
      from SiEPIC.extend import to_itype
      import math
      from pya import DPolygon
      # This is the main part of the implementation: create the layout

      # fetch the parameters
      dbu = self.layout.dbu
      ly = self.layout
      shapes = self.cell.shapes
      LayerSi = self.layer
      LayerSiN = ly.layer(LayerSi)
      LayerPinRecN = ly.layer(self.pinrec)
      LayerDevRecN = ly.layer(self.devrec)

      #variables
      pitch = self.pitch
      w = self.w
      r = self.r
      ff = self.ff
      angle = self.angle
      gap = self.gap
      W_ratio = self.W_ratio
      doublebus = self.doublebus
      if debug:
        print(doublebus)
      
      # set the deltaL of sT-shaped silicon pillar (deltaL = (Lin-Lout)/4) 
      deltaL = (2*pitch*ff)*(W_ratio-1)/((1+W_ratio)*(4))
      
      if (r-w/2 <= 0):
        r = 5
        if debug:
          print ('invalid radius, set r to default: 5')

      #Calculate number of segments
      s1 = pitch*ff #silicon
      s2 = pitch-s1 #gap
      
      #calculate best radius
      pi = math.pi
      const = math.ceil(2*pi*r/(s1+s2))
      #if doesn't divide evenly, replace r with best possible r
      if ((2*pi*r)%(s1+s2) != 0):
        r = const*(s1+s2)/(2*pi)
        if debug:
          print('r adjusted to '+str(r)+'um to fit periods perfectly.')
      
      #theta1 = (s1/r)
      #theta2 = (s2/r)
      theta1 = math.atan(s1/r) # angle of silicon compared to the origin point
      theta2 = math.atan(s2/r) # angle of the gap compared to the origin point
      nSeg = int(math.floor(angle/(math.degrees(theta1)+math.degrees(theta2)))) #how many segments to have
      si_first = True #for alternating between silicon and gap
      j=0 #index of how many silicon thetas 
      jj = 0 #index of how many gap thetas
      ORDER = True #ordering of the coordinates for polygon drawing
     
      #xo = [(r-w/2)*math.cos(0)]
      #yo = [(r-w/2)*math.sin(0)]
      #xo.append((r+w/2)*math.cos(0))
      #yo.append((r+w/2)*math.sin(0))
      
      xo = [(r-w/2)*math.cos(0)+deltaL*math.sin(0)]
      yo = [(r-w/2)*math.sin(0)-deltaL*math.cos(0)]
      xo.append((r+w/2)*math.cos(0)-deltaL*math.sin(0))
      yo.append((r+w/2)*math.sin(0) +deltaL*math.cos(0))
      
            
      for i in range(0,nSeg*2):
          if si_first:
              j = j+1
              si_first = not(si_first)
          else:
              jj=jj+1
              si_first = not(si_first)
          
          if ORDER: 
            xo.append((r+w/2)*math.cos(j*theta1+jj*theta2)+deltaL*math.sin(j*theta1+jj*theta2))
            yo.append((r+w/2)*math.sin(j*theta1+jj*theta2)-deltaL*math.cos(j*theta1+jj*theta2))
            xo.append((r-w/2)*math.cos(j*theta1+jj*theta2)-deltaL*math.sin(j*theta1+jj*theta2))
            yo.append((r-w/2)*math.sin(j*theta1+jj*theta2)+deltaL*math.cos(j*theta1+jj*theta2))
            ORDER = not(ORDER)
          else:
            xo.append((r-w/2)*math.cos(j*theta1+jj*theta2)+deltaL*math.sin(j*theta1+jj*theta2))
            yo.append((r-w/2)*math.sin(j*theta1+jj*theta2)-deltaL*math.cos(j*theta1+jj*theta2))
            xo.append((r+w/2)*math.cos(j*theta1+jj*theta2)-deltaL*math.sin(j*theta1+jj*theta2))
            yo.append((r+w/2)*math.sin(j*theta1+jj*theta2)+deltaL*math.cos(j*theta1+jj*theta2))
            ORDER = not(ORDER)    
        
          if len(xo) == 4:
            dpts=[pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
            shapes(LayerSiN).insert(element)
            xo = []
            yo = []  

      #Draw Bus WG
      #calulate ideal length of bus
      bus_length = self.cell.bbox().height()*dbu +pitch*2
      constant = math.ceil(bus_length/(s1+s2))
      if bus_length % (s1+s2) != 0:
        bus_length = constant*(s1+s2)
  
      #draw first box at center    
      xo = [(r+w/2+gap),(r+w/2+gap+w),(r+w/2+gap+w),(r+w/2+gap)]
      yo = [0,0,s1,s1]
      dpts=[pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
      dpolygon = DPolygon(dpts)
      element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
      shapes(LayerSiN).insert(element)
      
      for i in range(0,int(math.ceil((constant)/2))):
        yu = [yo[j]+i*pitch for j in range(len(yo))]
        yd = [yo[j]-i*pitch for j in range(len(yo))]
        
        dpts=[pya.DPoint(xo[i], yu[i]) for i in range(len(xo))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
        shapes(LayerSiN).insert(element)
        dpts=[pya.DPoint(xo[i], yd[i]) for i in range(len(xo))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
        shapes(LayerSiN).insert(element)   
        
      if doublebus:
        for i in range(0,int(math.ceil((constant)/2))):
          x2 = [xo[j]*-1 for j in range(len(xo))]
          yu = [yo[j]+i*pitch for j in range(len(yo))]
          yd = [yo[j]-i*pitch for j in range(len(yo))]
          
          dpts=[pya.DPoint(x2[i], yu[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)
          dpts=[pya.DPoint(x2[i], yd[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)   
          
     
      #DEV BOX
      if doublebus is False:
        half_l = (self.cell.bbox().width()-w/dbu-gap/dbu)/2
        half_r = self.cell.bbox().width()-half_l
        dev = Box(-half_l,yd[-4]/dbu,half_r,yu[-1]/dbu)
        shapes(LayerDevRecN).insert(dev)
        dev_width = self.cell.bbox().width()/2
      else:
        half_l = (self.cell.bbox().width())/2
        half_r = self.cell.bbox().width()-half_l
        dev = Box(-half_l,yd[-4]/dbu,half_r,yu[-1]/dbu)
        shapes(LayerDevRecN).insert(dev)
        dev_width = self.cell.bbox().width()/2
    
    # Create the pins on the waveguides, as short paths:
      from SiEPIC._globals import PIN_LENGTH as pin_length
    
      w = to_itype(self.w,dbu)
      gap = to_itype(self.gap,dbu)
      bus_length = to_itype(bus_length/2,dbu)
  
        #Pin1
      t = Trans(Trans.R0, half_r-w/2,yu[-1]/dbu)
      pin = Path([Point(0, -pin_length/2), Point(0, pin_length/2)], w)
      pin_t = pin.transformed(t)
      shapes(LayerPinRecN).insert(pin_t)
      text = Text ("pin1", t)
      shape = shapes(LayerPinRecN).insert(text)
      shape.text_size = 0.4/dbu
        
        #Pin2
      t = Trans(Trans.R0, half_r-w/2,yd[-4]/dbu)
      pin = Path([Point(0, pin_length/2), Point(0, -pin_length/2)], w)
      pin_t = pin.transformed(t)
      shapes(LayerPinRecN).insert(pin_t)
      text = Text ("pin2", t)
      shape = shapes(LayerPinRecN).insert(text)
      shape.text_size = 0.4/dbu
        
      if doublebus is True:
        half_l = self.cell.bbox().width()/2
        half_r = self.cell.bbox().width()-half_l
        #pin 3
        t = Trans(Trans.R0, -half_l+w/2,yu[-1]/dbu)
        pin = Path([Point(0, -pin_length/2), Point(0, pin_length/2)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text ("pin3", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4/dbu
        
        #Pin4
        t = Trans(Trans.R0, -half_l+w/2,yd[-4]/dbu)
        pin = Path([Point(0, pin_length/2), Point(0, -pin_length/2)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text ("pin4", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4/dbu




import pya
from pya import *
from SiEPIC.utils import get_technology, get_technology_by_name
from SiEPIC.utils import arc, arc_wg, arc_to_waveguide, points_per_circle#,layout
import math


class phc_H0c_new(pya.PCellDeclarationHelper):
  """
  Input: length, width
  """
  import numpy
  
  def __init__(self):

    # Important: initialize the super class
    super(phc_H0c_new, self).__init__()

    self.param("a", self.TypeDouble, "lattice constant (microns)", default = 0.744)     
    self.param("n", self.TypeInt, "Number of holes in x and y direction", default = 30)     
    self.param("r", self.TypeDouble, "hole radius (microns)", default = 0.179)
    self.param("wg_dis", self.TypeInt, "Waveguide distance (number of holes)", default = 3)        
    self.param("S1x", self.TypeDouble, "S1x shift", default = 0.28)     
    self.param("S2x", self.TypeDouble, "S2x shift", default = 0.193)     
    self.param("S3x", self.TypeDouble, "S3x shift", default = 0.194)     
    self.param("S4x", self.TypeDouble, "S4x shift", default = 0.162)
    self.param("S5x", self.TypeDouble, "S5x shift", default = 0.113)
    self.param("S1y", self.TypeDouble, "S1y shift", default = -0.016)
    self.param("S2y", self.TypeDouble, "S2y shift", default = 0.134)
    self.param("bus_number",  self.TypeInt, "2 for double, 1 for single, max 2", default = 2)
    TECHNOLOGY = get_technology_by_name('EBeam')
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = TECHNOLOGY['Text'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "H0 Cavity_a%s-r%.3f-wg_dis%.3f-n%.3f" % \
    (self.a, self.r, self.wg_dis, self.n)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    bus_n = self.bus_number
    
    LayerSi = self.layer
    LayerSiN = ly.layer(self.layer)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    LayerTextN = ly.layer(self.textl)
        
    a = self.a/dbu
    r = self.r/dbu
    wg_dis = self.wg_dis+1
    n = int(math.ceil(self.n/2))
    Sx = [self.S1x,self.S2x,self.S3x,self.S4x,self.S5x]
    Sy = [self.S1y,0,self.S2y]
    if wg_dis%2 == 0:
      length_slab_x = 2*n*a
    else:
      length_slab_x = (2*n+1)*a
    
    length_slab_y = 2*(n-2)*a

    if bus_n == 2:
      k = -1
    else:
      k = 1
    
    #function to creat polygon pts for right half of a hole in a hexagon unit cell
    def hexagon_hole_half(a,r): 
      npts = 10    
      theta_div = math.pi/3
      theta_div_hole = math.pi/npts
      triangle_length = a/math.sqrt(3)
      pts = []
      for i in range(0,4):
        pts.append(Point.from_dpoint(pya.DPoint(triangle_length*math.cos(i*theta_div-math.pi/2), triangle_length*math.sin(i*theta_div-math.pi/2))))
      for i in range(0, npts+1):
        pts.append(Point.from_dpoint(pya.DPoint(r*math.cos(math.pi/2-i*theta_div_hole), r*math.sin(math.pi/2-i*theta_div_hole))))
      return pts
    
    def hexagon_shifthole_half(a,r): 
      npts = 10    
      theta_div = math.pi/3
      theta_div_hole = math.pi/npts
      triangle_length = a*1.235/math.sqrt(3)
      pts = []
      for i in range(0,4):
        pts.append(Point.from_dpoint(pya.DPoint(triangle_length*math.cos(i*theta_div-math.pi/2), triangle_length*math.sin(i*theta_div-math.pi/2))))
      for i in range(0, npts+1):
        pts.append(Point.from_dpoint(pya.DPoint(r*math.cos(math.pi/2-i*theta_div_hole), r*math.sin(math.pi/2-i*theta_div_hole))))
      return pts
    
    #function to creat polygon pts for right half of a hexagon unit cell      
    def hexagon_half(a): 
      theta_div = math.pi/3
      triangle_length = a/math.sqrt(3)
      pts = []
      for i in range(0,4):
        pts.append(Point.from_dpoint(pya.DPoint(triangle_length*math.cos(i*theta_div-math.pi/2), triangle_length*math.sin(i*theta_div-math.pi/2))))
      return pts      
    
    #create the right and left half of the hole and hexagon cells
    #hole_cell = pya.Region()
    #hexagon_cell = pya.Region()
    hole = pya.Region()
    
    hole_cell_pts = hexagon_hole_half(a,r)
    hexagon_pts = hexagon_half(a)
    hole_shiftcell_pts = hexagon_shifthole_half(a,r)
    hole_cell_poly_0 = pya.Polygon(hole_cell_pts)
    hexagon_cell_poly_0 = pya.Polygon(hexagon_pts)
    hole_shiftcell_poly_0 = pya.Polygon(hole_shiftcell_pts)

    hole_trans = pya.Trans(pya.Trans.R180)
    hole_cell_poly_1 = hole_cell_poly_0.transformed(hole_trans)
    hexagon_cell_poly_1 = hexagon_cell_poly_0.transformed(hole_trans)    
    hole_shiftcell_poly_1 = hole_shiftcell_poly_0.transformed(hole_trans)
        
    #create the photonic crystal with shifts and waveguides
    for j in range(-n+1,n):
      if j%2 == 0:
        for i in range(-n,n+1):
          #waveguide          
          if (j == k*wg_dis and i > 3) or (j == wg_dis and i != 0):
            hole_x = abs(i)/i*(abs(i)-0.5)*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hexagon_cell_poly_0.transformed(hole_trans)
            hole_t_1 = hexagon_cell_poly_1.transformed(hole_trans)            
            hole.insert(hole_t_0)
            hole.insert(hole_t_1)
          #filling the edges with half cell
          elif i in (-n,n) and wg_dis%2 == 1:
            hole_x = abs(i)/i*(abs(i)+0.5)*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            if i == -n:         
              hole_t = hole_cell_poly_0.transformed(hole_trans)
            else:
              hole_t = hole_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t)
            hole_x = abs(i)/i*(abs(i)-0.5)*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
            hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t_0)  
            hole.insert(hole_t_1)             
          #x shifts
          elif j == 0 and i in (1,-1,2,-2,3,-3,4,-4,5,-5):
            hole_x = abs(i)/i*(abs(i)-0.5+Sx[abs(i)-1])*a
            hole_y = 0 
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hole_shiftcell_poly_0.transformed(hole_trans)
            hole_t_1 = hole_shiftcell_poly_1.transformed(hole_trans)
            hole.insert(hole_t_0)  
            hole.insert(hole_t_1)                    
          elif i!=0:
            hole_x = abs(i)/i*(abs(i)-0.5)*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
            hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t_0)  
            hole.insert(hole_t_1)
      elif j%2 == 1:
        for i in range(-n,n+1):
          #waveguide
          if (j == k*wg_dis and i > 3) or j == wg_dis:
            hole_x = i*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hexagon_cell_poly_0.transformed(hole_trans)
            hole_t_1 = hexagon_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t_0)
            hole.insert(hole_t_1)
          #filling the edges with half cell
          elif wg_dis%2 == 0 and i in (-n,n):
            hole_x = i*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y) 
            if i == -n:           
              hole_t = hole_cell_poly_0.transformed(hole_trans)
            else:
              hole_t = hole_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t)               
          #y shifts
          elif i == 0 and j in (1,-1,3,-3):
            hole_x = 0
            hole_y = j*a*(math.sqrt(3)/2)+abs(j)/j*a*Sy[abs(j)-1]
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hole_shiftcell_poly_0.transformed(hole_trans)
            hole_t_1 = hole_shiftcell_poly_1.transformed(hole_trans)
            hole.insert(hole_t_0)  
            hole.insert(hole_t_1)
          else:  
            hole_x = i*a
            hole_y = j*a*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
            hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
            hole.insert(hole_t_0)  
            hole.insert(hole_t_1)
    
    #print(hole_t_0)
    box_l = a/2
    hole.insert(pya.Box(-box_l,-box_l,box_l,box_l))
    cover_box = pya.Box(-length_slab_x/2, -a/2, length_slab_x/2, a/2) 
    box_y = n*a*math.sqrt(3)/2  
    cover_box_trans_0 = pya.Trans(Trans.R0, 0,box_y)
    cover_box_trans_1 = pya.Trans(Trans.R0, 0,-box_y) 
    cover_box_t_0 = cover_box.transformed(cover_box_trans_0) 
    cover_box_t_1 = cover_box.transformed(cover_box_trans_1)     
    #hole.insert(pya.Box())    
    self.cell.shapes(LayerSiN).insert(hole)
    self.cell.shapes(LayerSiN).insert(cover_box_t_0)
    self.cell.shapes(LayerSiN).insert(cover_box_t_1)      
    # Pins on the waveguide:    
    pin_length = 200
    pin_w = a
    wg_pos = a*math.sqrt(3)/2*wg_dis
    
    t = pya.Trans(Trans.R0, -length_slab_x/2,wg_pos)
    pin = pya.Path([pya.Point(-pin_length/2, 0), pya.Point(pin_length/2, 0)], pin_w)
    pin_t = pin.transformed(t)
    self.cell.shapes(LayerPinRecN).insert(pin_t)
    text = pya.Text ("pin1", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    t = pya.Trans(Trans.R0, length_slab_x/2,wg_pos)
    pin_t = pin.transformed(t)
    self.cell.shapes(LayerPinRecN).insert(pin_t)
    text = pya.Text ("pin2", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    
    #pin for drop waveguide
    t = pya.Trans(Trans.R0, length_slab_x/2,-wg_pos)
    pin_t = pin.transformed(t)
    self.cell.shapes(LayerPinRecN).insert(pin_t)
    text = pya.Text ("pin3", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    points = [[-length_slab_x/2,0], [length_slab_x/2, 0]]
    points = [Point(each[0], each[1]) for each in points]
    path = Path(points, length_slab_y)   
    self.cell.shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
from . import *
from SiEPIC.utils import get_technology_by_name
from SiEPIC._globals import PIN_LENGTH as pin_length
from pya import DPolygon
import math

#====================================================================================================
# File:     e_skid.py
# Author:   Hang (Bobby) Zou
# Purpose:  PCell in EBeam library for a common parallel e-skid design that have adjustable period
#           and fill factor
# Version:  V2.0
# Date:     2023-11-09
#====================================================================================================

class e_skid(pya.PCellDeclarationHelper):

    def __init__(self):

        # Important: initialize the super class
        super(e_skid, self).__init__()
        self.technology_name = 'EBeam'

        TECHNOLOGY = get_technology_by_name(self.technology_name)

        # e_skid Parameters:
        # core parameter
        self.param('divider0',      self.TypeNone,    '==============Core Parameter==============')
        
        self.param('w',             self.TypeDouble,  'Core Waveguide Width [um]'           , default = 0.3)
        
        self.param('divider1',      self.TypeNone,    '==============Clad Parameter==============')

        # clad parameter
        self.param('p',             self.TypeDouble,  'e-skid Period [um]'                  , default = 0.1)
        self.param('ff',            self.TypeDouble,  'e-skid Fill Factor [0.2-0.8]'        , default = 0.5)
        self.param('p_num',         self.TypeInt,     'e-skid Period Number'                , default = 5)
        self.param('length',        self.TypeDouble,  'Length of e-skid Waveguide [um]'     , default = 10)

        self.param('divider2',      self.TypeNone,    '==============Taper Parameter=============')

        # taper
        self.param('taper_True',    self.TypeBoolean, 'Taper [T/F]'                         , default = True)
        self.param('taperL',        self.TypeDouble,  'Strip -> e-skid Taper Length [um]'   , default = 5)
        self.param('taperW',        self.TypeDouble,  'Strip Width [um]'                    , default = 0.5)

        self.param('divider3',      self.TypeNone,    '============Extended Clad Parameter==========')        

        # extended clad parameter
        self.param('clad_taper_L',  self.TypeDouble,  'Extended e-skid Length [um]'         , default = 5)
        self.param('clad_taper_angle', self.TypeDouble,'Extended e-skid Angle [degree]'     , default = 2)
        
        self.param('divider4',      self.TypeNone,    '=====================================')

        # Layer Parameters - Don't touch
        self.param("layer",  self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
        self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
        self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
        self.param("oxideopen", self.TypeLayer, "Oxide Open Layer", default = TECHNOLOGY['Oxide open (to BOX)'])
    
    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "configurable_e_skid"

    # Creating the layout
    def produce_impl(self):
        
        # Layout Parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        w = self.w                              # Core Waveguide Width [um]
        p = self.p                              # e-skid Period [um]
        ff = self.ff                            # e-skid Duty Cycle [%]
        p_num = self.p_num                      # e-skid Period Number
        length = self.length                    # e-skid Waveguide Length [um]
        
        taper_True = self.taper_True            # e-skid Taper [T/F]
        taperW = self.taperW                    # e-skid Taper Width [um]
        taperL = self.taperL                    # e-skid Taper Length [um]

        clad_taper_L = self.clad_taper_L        # Extended e-skid length [nm]
        clad_taper_angle_radians = self.clad_taper_angle*math.pi/180    # Extended e-skid angle [degrees]
        
        

        ###############################
        ########### e-skid ############
        ###############################
        
        # 1. e-skid main body

        # x, y coordinates of the core
        x_core = [0, length, length, 0]              
        y_core = [-w/2, - w/2, w/2, w/2]
        
        dpts = [pya.DPoint(x_core[i], y_core[i]) for i in range(len(x_core))] # Core body
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon*(1/dbu))
        shapes(LayerSiN).insert(element)

        # draw multiple clad for both top and bottom
        for i in range (1, p_num+1):
        
            # x, y coordinates of the top clad
            x_clad_top = [0, length, length, 0] 
            y_clad_top = [w/2+i*p-(p*ff), w/2+i*p-(p*ff), w/2+i*p, w/2+i*p]

            dpts = [pya.DPoint(x_clad_top[i], y_clad_top[i]) for i in range(len(x_clad_top))] # Core body
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon*(1/dbu))
            shapes(LayerSiN).insert(element)

            # x, y coordinates of the bottom clad
            x_clad_bottom = [0, length, length, 0] 
            y_clad_bottom = [-w/2-i*p+(p*ff), -w/2-i*p+(p*ff), -w/2-i*p, -w/2-i*p]

            dpts = [pya.DPoint(x_clad_bottom[i], y_clad_bottom[i]) for i in range(len(x_clad_bottom))] # Core body
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon*(1/dbu))
            shapes(LayerSiN).insert(element)

        # 2. taper + Pins

        if taper_True:
            # Top clad
            if clad_taper_L > taperL:
                # Prevent the clad_taper_L to be larger than taperL
                clad_taper_L = taperL

            # Calculate the required offset
            w_offset = taperW * clad_taper_L / taperL

            # fetch the taper angle at the set point
            taper_angle = math.atan(((taperW-w)/2)/(taperL))
            
            if clad_taper_angle_radians <= taper_angle:
                clad_taper_angle_radians = taper_angle


            for i in range (1, p_num+1):
    
                # draw the right clad, angle method
                # locate the end points of the extended clad
                clad_end_point_position = w/2+i*p + i*(clad_taper_L*math.tan(clad_taper_angle_radians))
                

                # x, y coordinates of the top clad
                x_clad_top = [length, length+clad_taper_L, length+clad_taper_L, length]
                y_clad_top = [w/2+i*p-(p*ff) , clad_end_point_position-(p*ff),clad_end_point_position, w/2+i*p]

                dpts = [pya.DPoint(x_clad_top[i], y_clad_top[i]) for i in range(len(x_clad_top))] # Core body
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon*(1/dbu))
                shapes(LayerSiN).insert(element)

                # x, y coordinates of the bottom clad
                x_clad_bottom = [length, length+clad_taper_L, length+clad_taper_L, length] 
                y_clad_bottom = [-w/2-i*p+(p*ff) , -clad_end_point_position+(p*ff),-clad_end_point_position, -w/2-i*p]

                dpts = [pya.DPoint(x_clad_bottom[i], y_clad_bottom[i]) for i in range(len(x_clad_bottom))] # Core body
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon*(1/dbu))
                shapes(LayerSiN).insert(element)


                # draw the left clad, angle method
                # locate the end points of the extended clad
                clad_end_point_position = w/2+i*p + i*(clad_taper_L*math.tan(clad_taper_angle_radians))
                
                # x, y coordinates of the top clad
                x_clad_top = [0, -clad_taper_L, -clad_taper_L, 0]
                y_clad_top = [w/2+i*p-(p*ff) , clad_end_point_position-(p*ff),clad_end_point_position, w/2+i*p]

                dpts = [pya.DPoint(x_clad_top[i], y_clad_top[i]) for i in range(len(x_clad_top))] # Core body
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon*(1/dbu))
                shapes(LayerSiN).insert(element)

                # x, y coordinates of the bottom clad
                x_clad_bottom = [0, -clad_taper_L, -clad_taper_L, 0] 
                y_clad_bottom = [-w/2-i*p+(p*ff) , -clad_end_point_position+(p*ff),-clad_end_point_position, -w/2-i*p]

                dpts = [pya.DPoint(x_clad_bottom[i], y_clad_bottom[i]) for i in range(len(x_clad_bottom))] # Core body
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon*(1/dbu))
                shapes(LayerSiN).insert(element)
            
            # left taper
            x_taper_left = [0, 0, -taperL, -taperL]              
            y_taper_left = [-w/2, w/2, taperW/2, -taperW/2]
            
            dpts = [pya.DPoint(x_taper_left[i], y_taper_left[i]) for i in range(len(x_taper_left))] # Core body
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon*(1/dbu))
            shapes(LayerSiN).insert(element)

            # right taper
            x_taper_right = [length, length, length+taperL, length+taperL]              
            y_taper_right = [-w/2, w/2, taperW/2, -taperW/2]
            
            dpts = [pya.DPoint(x_taper_right[i], y_taper_right[i]) for i in range(len(x_taper_right))] # Core body
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon*(1/dbu))
            shapes(LayerSiN).insert(element)

            # Adding Pins for mating
            # Pin 1
            t = pya.DTrans(pya.Trans.R90, -taperL/dbu, 0/dbu)
            pin = pya.Path([pya.DPoint(0, -pin_length/2), pya.DPoint(0, pin_length/2)], taperW/dbu)
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = pya.Text ("opt1", pya.DTrans(pya.Trans.R0, -taperL/dbu, 0/dbu))
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.2/dbu

            # Pin 2
            t = pya.DTrans(pya.Trans.R270, (taperL + length)/dbu, 0/dbu)
            pin = pya.Path([pya.DPoint(0, -pin_length/2), pya.DPoint(0, pin_length/2)], taperW/dbu)
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = pya.Text ("opt2", pya.DTrans(pya.Trans.R0, (taperL + length)/dbu, 0/dbu))
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.2/dbu

            # Devbox 
            dev = pya.DBox(-taperL, 
                            clad_end_point_position + 0.5, 
                            length + taperL,
                            -clad_end_point_position - 0.5)
                        
            shapes(LayerDevRecN).insert(dev)

        else:
            # Adding Pins for mating
            # Pin 1
            t = pya.DTrans(pya.Trans.R90, 0, 0/dbu)
            pin = pya.Path([pya.DPoint(0, -pin_length/2), pya.DPoint(0, pin_length/2)], w/dbu)
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = pya.Text ("opt1", pya.DTrans(pya.Trans.R0, 0, 0/dbu))
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.2/dbu

            # Pin 2
            t = pya.DTrans(pya.Trans.R270, length/dbu, 0/dbu)
            pin = pya.Path([pya.DPoint(0, -pin_length/2), pya.DPoint(0, pin_length/2)], w/dbu)
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = pya.Text ("opt2", pya.DTrans(pya.Trans.R0, length/dbu, 0/dbu))
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.2/dbu

            # Devbox 
            dev = pya.DBox(0, 
                            w/2 + p_num*p + 0.5, 
                            length,
                            -w/2 - p_num*p - 0.5)
                        
            shapes(LayerDevRecN).insert(dev)

#====================================================================================================
# END OF DOCUMENT
#====================================================================================================


"""
This file is part of the SiEPIC-EBeam-PDK

Version history:

Lukas Chrostowski 2022/04/01
 - PCell takes an input of the waveguide type from a dropdown list, rather than 
   detailed parameters.  Loaded from Waveguides.xml
"""

from pya import *
import pya

class Waveguide(pya.PCellDeclarationHelper):

  def __init__(self):
    # Important: initialize the super class
    super(Waveguide, self).__init__()
    
    from SiEPIC.utils import get_technology_by_name, load_Waveguides_by_Tech

    '''
    # https://github.com/KLayout/klayout/issues/879
    tech = self.layout.library().technology
    if not tech:
       tech = 'EBeam'
    self.technology_name = tech
    '''
    self.technology_name = 'EBeam'
            
    # Load all strip waveguides
    self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)   
        
    # declare the parameters

    p = self.param("waveguide_type", self.TypeList, "Waveguide Type", default = self.waveguide_types[0]['name'])
    for wa in self.waveguide_types:
        p.add_choice(wa['name'],wa['name'])
    self.param("path", self.TypeShape, "Path", default = DPath([DPoint(0,0), DPoint(10,0), DPoint(10,10)], 0.5))

    # self.param("length", self.TypeDouble, "Length", default = 0, readonly=True)
    ''' todo - can add calculated values and parameters for user info
    self.param("radius", self.TypeDouble, "Bend Radius", default = 0, readonly=True)
    '''
    
    self.cellName="Waveguide"

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    if 'waveguide_length' in dir(self):
        return "%s_%s_%.3f" % (self.cellName, self.waveguide_type.replace(' ','_').replace(',',''), self.waveguide_length)
    else:
        return "%s_%s_%.3f" % (self.cellName, self.waveguide_type.replace(' ','_').replace(',',''), 0)
  
  def coerce_parameters_impl(self):
#    self.length = self.waveguide_length
    pass                  
  def can_create_from_shape_impl(self):
    return self.shape.is_path()

  def transformation_from_shape_impl(self):
    return Trans(Trans.R0,0,0)

  def parameters_from_shape_impl(self):
    self.path = self.shape.path
        
  def produce_impl(self):

    # https://github.com/KLayout/klayout/issues/879
    # tech = self.layout.library().technology
        
    # Make sure the technology name is associated with the layout
    #  PCells don't seem to know to whom they belong!
    if self.layout.technology_name == '':
        self.layout.technology_name = self.technology_name

    # Draw the waveguide geometry, new function in SiEPIC-Tools v0.3.90
    from SiEPIC.utils.layout import layout_waveguide4
    self.waveguide_length = layout_waveguide4(self.cell, self.path, self.waveguide_type, debug=False)

    # print("EBeam.%s: length %.3f um, complete" % (self.cellName, self.waveguide_length))

'''
import pya
from pya import *

class Waveguide(pya.PCellDeclarationHelper):

  def __init__(self):
    # Important: initialize the super class
    super(Waveguide, self).__init__()
    # declare the parameters
    from SiEPIC.utils import get_technology_by_name
    TECHNOLOGY = get_technology_by_name('EBeam')
    self.param("path", self.TypeShape, "Path", default = DPath([DPoint(0,0), DPoint(10,0), DPoint(10,10)], 0.5))
    self.param("radius", self.TypeDouble, "Radius", default = 5)
    self.param("width", self.TypeDouble, "Width", default = 0.5)
    self.param("adiab", self.TypeBoolean, "Adiabatic", default = False)
    self.param("bezier", self.TypeDouble, "Bezier Parameter", default = 0.35)
    self.param("layers", self.TypeList, "Layers", default = ['Waveguide'])
    self.param("widths", self.TypeList, "Widths", default =  [0.5])
    self.param("offsets", self.TypeList, "Offsets", default = [0])
    self.param("CML", self.TypeString, "Compact Model Library (CML)", default = 'EBeam')
    self.param("model", self.TypeString, "CML Model name", default = 'ebeam_wg_integral_1550') 
    self.cellName="Waveguide"
    
  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "%s_%s" % (self.cellName, self.path)
  
  def coerce_parameters_impl(self):
    from SiEPIC.extend import to_itype
    print("EBeam.Waveguide coerce parameters")
    
    if 0:
        TECHNOLOGY = get_technology_by_name('EBeam')
        dbu = self.layout.dbu
        wg_width = to_itype(self.width,dbu)
        for lr in range(0, len(self.layers)):
          layer = self.layout.layer(TECHNOLOGY[self.layers[lr]])
          width = to_itype(self.widths[lr],dbu)
          # check to make sure that the waveguide with parameters are consistent in both places
          if self.layout.layer(TECHNOLOGY['Waveguide']) == layer:
            if width != wg_width:
              self.widths[lr] = self.width
          # check to make sure that the DevRec is bigger than the waveguide width
          if self.layout.layer(TECHNOLOGY['DevRec']) == layer:
            if width < wg_width:
              self.widths[lr] = self.width*2
              
  def can_create_from_shape_impl(self):
    return self.shape.is_path()

  def transformation_from_shape_impl(self):
    return Trans(Trans.R0,0,0)

  def parameters_from_shape_impl(self):
    self.path = self.shape.dpath
        
  def produce_impl(self):

    from SiEPIC.utils.layout import layout_waveguide2
    from SiEPIC.utils import  angle_vector
    from math import cos, sin, pi, sqrt
    import pya
    from SiEPIC.extend import to_itype
    
    # print("EBeam.Waveguide")
    
    from SiEPIC.utils import get_technology_by_name
    TECHNOLOGY = get_technology_by_name('EBeam')
    
    dbu = self.layout.dbu
    wg_width = to_itype(self.width,dbu)
    path = self.path.to_itype(dbu)

    if not (len(self.layers)==len(self.widths) and len(self.layers)==len(self.offsets) and len(self.offsets)==len(self.widths)):
      raise Exception("There must be an equal number of layers, widths and offsets")
    path.unique_points()
    pts = path.get_points()
    
    # Draw the waveguide geometry, new in SiEPIC-Tools v0.3.64
    waveguide_length = layout_waveguide2(TECHNOLOGY, self.layout, self.cell, self.layers, self.widths, self.offsets, pts, self.radius, self.adiab, self.bezier)

    pts = path.get_points()
    LayerPinRecN = self.layout.layer(TECHNOLOGY['PinRec'])
    
    t1 = Trans(angle_vector(pts[0]-pts[1])/90, False, pts[0])
    self.cell.shapes(LayerPinRecN).insert(Path([Point(-5, 0), Point(5, 0)], wg_width).transformed(t1))
    self.cell.shapes(LayerPinRecN).insert(Text("pin1", t1, 0.3/dbu, -1))
    
    t = Trans(angle_vector(pts[-1]-pts[-2])/90, False, pts[-1])
    self.cell.shapes(LayerPinRecN).insert(Path([Point(-5, 0), Point(5, 0)], wg_width).transformed(t))
    self.cell.shapes(LayerPinRecN).insert(Text("pin2", t, 0.3/dbu, -1))

    LayerDevRecN = self.layout.layer(TECHNOLOGY['DevRec'])

    # Compact model information
    angle_vec = angle_vector(pts[0]-pts[1])/90
    halign = 0 # left
    angle=0
    if angle_vec == 0: # horizontal
      halign = 2 # right
      angle=0
      dpt = Point(0, 0.2*wg_width)
    if angle_vec == 2: # horizontal
      halign = 0 # left
      angle = 0
      dpt=Point(0, 0.2*wg_width)
    if angle_vec == 1: # vertical
      halign = 2 # right
      angle = 1
      dpt=Point(0.2*wg_width,0)
    if angle_vec == -1: # vertical
      halign = 0 # left
      angle = 1
      dpt=Point(0.2*wg_width,0)
    pt2=pts[0] + dpt
    pt3=pts[0] - dpt
    pt4=pts[0] - 2*dpt
    pt5=pts[0] + 2*dpt
      

    t = Trans(angle, False, pt5)
    text = Text ('cellName=%s' % self.cellName, t, 0.1*wg_width, -1)
    text.halign=halign
    shape = self.cell.shapes(LayerDevRecN).insert(text)
    t = Trans(angle, False, pt3) 
    text = Text ('Lumerical_INTERCONNECT_library=Design kits/%s' % self.CML, t, 0.1*wg_width, -1)
    text.halign=halign
    shape = self.cell.shapes(LayerDevRecN).insert(text)
    t = Trans(angle, False, pt2)
    text = Text ('Component=%s' % self.model, t, 0.1*wg_width, -1)
    text.halign=halign
    shape = self.cell.shapes(LayerDevRecN).insert(text)
    t = Trans(angle, False, pts[0])
    pts_txt = str([ [round(p.to_dtype(dbu).x,3), round(p.to_dtype(dbu).y,3)] for p in pts ]).replace(', ',',')
    text = Text ( \
      'Spice_param:wg_length=%.3fu wg_width=%.3fu points="%s" radius=%s' %\
        (waveguide_length, self.width, pts_txt,self.radius ), t, 0.1*wg_width, -1  )
    text.halign=halign
    shape = self.cell.shapes(LayerDevRecN).insert(text)
    t = Trans(angle, False, pt4)
    text = Text ( \
      'Length=%.3fu' %(waveguide_length), t, 0.5*wg_width, -1  )
    text.halign=halign
    shape = self.cell.shapes(LayerDevRecN).insert(text)
'''


import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class Contra_DC_SWG(pya.PCellDeclarationHelper):
  """
  Author:   Mustafa Hammood 
            Mustafa@siepic.com
  """

  def __init__(self):

    # Important: initialize the super class
    super(Contra_DC_SWG, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("number_of_periods", self.TypeInt, "Number of grating periods", default = 300)     
    self.param("grating_period", self.TypeDouble, "Grating period (microns)", default = 0.24)
    self.param("gap", self.TypeDouble, "Minimum gap (microns)", default = 0.15)          
    self.param("corrugation_width1", self.TypeDouble, "Waveguide 1 Corrugration width (microns)", default = 0.03)
    self.param("corrugation_width2", self.TypeDouble, "Waveguide 2 Corrugration width (microns)", default = 0.04) 
    self.param("wg1_width", self.TypeDouble, "Waveguide 1 width", default = 0.45)
    self.param("wg2_width", self.TypeDouble, "Waveguide 2 width", default = 0.55)
    self.param("duty", self.TypeDouble, "Duty cycle (0-1)", default = 0.5) 
    self.param("a", self.TypeDouble, "Gaussian Index", default = 2.7)   
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
#    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Contra_DC_SWG%s-%.3f" % \
    (self.number_of_periods, self.grating_period)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    import math
    
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

    from SiEPIC.extend import to_itype
    
    N = self.number_of_periods
    grating_period = int(round(self.grating_period/dbu))
    misalignment = 0
     
    # Determine the period such that the waveguide length is as desired.  Slight adjustment to period
    N_boxes = N
    
    # Draw the Bragg grating:
    box_width = int(round(grating_period*self.duty))
    
    w1 = self.wg1_width / dbu
    half_w1 = w1/2
    w2 = self.wg2_width / dbu
    half_w2 = w2/2
    deltaW1_max = to_itype(self.corrugation_width1,dbu)
    deltaW2_max = to_itype(self.corrugation_width2,dbu)


          
    for i in range(0,N_boxes+1):

      # apply apodization
      profileFunction = math.exp( -0.5*(2*self.a*(i-N/2)/(N))**2 )
      deltaW1 = deltaW1_max*profileFunction
      deltaW2 = deltaW2_max*profileFunction
      
      vertical_offset = int(round(self.wg2_width/2/dbu))+int(round(self.gap/dbu))+int(round(self.wg1_width/2/dbu))#+(-int(round(deltaW1))+int(round(deltaW2)))/2
      t = Trans(Trans.R0, 0,vertical_offset)
      
      if i%2 == True:
        x = int(round((i * grating_period - box_width/2)))
        box1_a = Box(x, -half_w1-deltaW1, x + box_width, half_w1-deltaW1)
        shapes(LayerSiN).insert(box1_a)
        
        box2_a = Box(x+grating_period, -half_w2-deltaW2, x + grating_period+box_width, half_w2-deltaW2).transformed(t)
        shapes(LayerSiN).insert(box2_a)
        
      else:
        x = int(round((i * grating_period - box_width/2)))
        box1_b = Box(x, -half_w1, x + box_width, half_w1)
        shapes(LayerSiN).insert(box1_b)
        
        box2_b = Box(x+grating_period, -half_w2, x +grating_period+ box_width, half_w2).transformed(t)
        shapes(LayerSiN).insert(box2_b)
      
    # missing periods due to misalignments
    box_final = Box(x+grating_period, -half_w1, x +grating_period+ box_width, half_w1)
    shapes(LayerSiN).insert(box_final)
    box_final = Box(-box_width/2, -half_w2-deltaW2, box_width/2, half_w2-deltaW2).transformed(t)
    shapes(LayerSiN).insert(box_final)
    
    # Create the pins on the waveguides, as short paths:
    from SiEPIC._globals import PIN_LENGTH as pin_length

    w = to_itype(self.wg1_width,dbu)
    t = Trans(Trans.R0, to_itype(-box_width/2,dbu*1000),-deltaW1/2)
    pin = Path([Point(pin_length/2, 0), Point(-pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    
    w = to_itype(self.wg2_width,dbu)
    t = Trans(Trans.R0, to_itype(-box_width/2,dbu*1000),vertical_offset-deltaW2/2)
    pin = Path([Point(pin_length/2, 0), Point(-pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    
    w = to_itype(self.wg2_width,dbu)
    t = Trans(Trans.R0, to_itype(x+grating_period+box_width,dbu*1000),vertical_offset-deltaW2/2)
    pin = Path([Point(-pin_length/2, 0), Point(pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin3", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    w = to_itype(self.wg1_width,dbu)
    t = Trans(Trans.R0, to_itype(x+grating_period+box_width,dbu*1000),-deltaW1/2)
    pin = Path([Point(-pin_length/2, 0), Point(pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin4", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
from . import *

class MMI_2x2(pya.PCellDeclarationHelper):
  """
  """
  def __init__(self):

    # Important: initialize the super class
    super(MMI_2x2, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("w", self.TypeDouble, "Waveguide Width", default = 0.5)
    self.param("w_mmi", self.TypeDouble, "Waveguide Width", default = 6.0)
    self.param("w_a", self.TypeDouble, "Access waveguide Width", default = 1.5)
    self.param("L_mmi", self.TypeDouble, "MMI Length", default = 130.5)
    self.param("L_a", self.TypeDouble, "Access waveguide Length", default = 10.0)
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "MMI_2x2"

  def can_create_from_shape_impl(self):
    return False
    
  def produce_impl(self):
    # This is the main part of the implementation: create the layout

    # test by Lukas, to see if it breaks the PCell
    # print(1/0)
    
    from math import pi, cos, sin

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes
    
    LayerSi = self.silayer
    LayerSiN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    TextLayerN = ly.layer(self.textl)

    
    w = int(round(self.w/dbu))
    w_mmi = int(round(self.w_mmi/dbu))
    w_a = int(round(self.w_a/dbu))
    L_mmi = int(round(self.L_mmi/dbu))
    L_a = int(round(self.L_a/dbu))

    # Create the MMI
    mmi = Box(-L_mmi/2, -w_mmi/2, L_mmi/2, w_mmi/2)
    shapes(LayerSiN).insert(mmi)

    # Draw taper access waveguides
    # Create a sub-cell
    cell = self.cell.layout().create_cell("MMI_access_waveguide")
    pts = []
    pts.append(Point.from_dpoint(DPoint(0, w_a/2)))
    pts.append(Point.from_dpoint(DPoint(L_a, w/2)))
    pts.append(Point.from_dpoint(DPoint(L_a, -w/2)))
    pts.append(Point.from_dpoint(DPoint(0, -w_a/2)))
    polygon = Polygon(pts)
    cell.shapes(LayerSiN).insert(polygon)
    # place the cell in the top cell
    self.cell.insert(CellInstArray(cell.cell_index(), Trans(Trans.R0, L_mmi/2, w_a)))
    self.cell.insert(CellInstArray(cell.cell_index(), Trans(Trans.R0, L_mmi/2, -w_a)))
    self.cell.insert(CellInstArray(cell.cell_index(), Trans(Trans.R180, -L_mmi/2, w_a)))
    self.cell.insert(CellInstArray(cell.cell_index(), Trans(Trans.R180, -L_mmi/2, -w_a)))


    pin(500,"pin1", Trans(Trans.R180, -L_mmi/2-L_a, w_a), LayerPinRecN, dbu, self.cell)
    pin(500,"pin2", Trans(Trans.R180, -L_mmi/2-L_a, -w_a), LayerPinRecN, dbu, self.cell)
    pin(500,"pin3", Trans(Trans.R0, L_mmi/2+L_a, w_a), LayerPinRecN, dbu, self.cell)
    pin(500,"pin4", Trans(Trans.R0, L_mmi/2+L_a, -w_a), LayerPinRecN, dbu, self.cell)

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    dev = Box(-L_mmi/2-L_a, w_mmi/2+w, L_mmi/2+L_a, -w_mmi/2-w )
    shapes(LayerDevRecN).insert(dev)

    # print("Done drawing the layout for - mmi" )



import pya
from pya import *

class ebeam_taper_te1550(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the strip waveguide taper.
  """

  def __init__(self):

    # Important: initialize the super class
    super(ebeam_taper_te1550, self).__init__()
    from SiEPIC.utils import get_technology_by_name
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("wg_width1", self.TypeDouble, "Waveguide Width1 (CML only supports 0.4, 0.5, 0.6)", default = 0.5)
    self.param("wg_width2", self.TypeDouble, "Waveguide Width2 (CML only supports 1, 2, 3)", default = 3)
    self.param("wg_length", self.TypeDouble, "Waveguide Length (CML only supports a range of 1-10)", default = 10)
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    # hidden parameters, can be used to query this component:
    self.param("p1", self.TypeShape, "DPoint location of pin1", default = Point(-10000, 0), hidden = True, readonly = True)
    self.param("p2", self.TypeShape, "DPoint location of pin2", default = Point(0, 10000), hidden = True, readonly = True)
    

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "ebeam_taper_te1550(R=" + ('%.3f-%.3f-%.3f' % (self.wg_width1,self.wg_width2,self.wg_length) ) + ")"

  def can_create_from_shape_impl(self):
    return False


  def produce(self, layout, layers, parameters, cell):
    """
    coerce parameters (make consistent)
    """
    self._layers = layers
    self.cell = cell
    self._param_values = parameters
    self.layout = layout
    shapes = self.cell.shapes
    
    from SiEPIC.extend import to_itype


    # cell: layout cell to place the layout
    # LayerSiN: which layer to use
    # w: waveguide width
    # length units in dbu

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    
    LayerSi = self.silayer
    LayerSiN = self.silayer_layer
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    
    w1 = to_itype(self.wg_width1,dbu)
    w2 = to_itype(self.wg_width2,dbu)
    length = to_itype(self.wg_length,dbu)

    pts = [Point(0,-w1/2), Point(0,w1/2), Point(length,w2/2), Point(length,-w2/2)]
    shapes(LayerSiN).insert(Polygon(pts))

    
    # Create the pins on the waveguides, as short paths:
    from SiEPIC._globals import PIN_LENGTH as pin_length
    
    # Pin on the left side:
    p1 = [Point(pin_length/2,0), Point(-pin_length/2,0)]
    p1c = Point(0,0)
    self.set_p1 = p1c
    self.p1 = p1c
    pin = Path(p1, w1)
    shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, 0, 0)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = to_itype(0.4,dbu)

    # Pin on the right side:
    p2 = [Point(length-pin_length/2,0), Point(length+pin_length/2,0)]
    p2c = Point(length, 0)
    self.set_p2 = p2c
    self.p2 = p2c
    pin = Path(p2, w2)
    shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, length, 0)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size =to_itype(0.4,dbu)
    shape.text_halign = 2

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    w_max = max([w1,w2])
    w_min = max([w1,w2])
    path = Path([Point(0,0),Point(length,0)],(w_max+w_min)*2)
    shapes(LayerDevRecN).insert(path.simple_polygon())


    # Compact model information
    t = Trans(Trans.R0, w1/10, 0)
    text = Text ("Lumerical_INTERCONNECT_library=Design kits/ebeam", t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = length/100
    t = Trans(Trans.R0, length/10, w1/4)
    text = Text ('Component=ebeam_taper_te1550', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = length/100
    t = Trans(Trans.R0, length/10, w1/2)
    text = Text ('Spice_param:wg_width1=%.3fu wg_width2=%.3fu wg_length=%.3fu'% (self.wg_width1,self.wg_width2,self.wg_length), t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = length/100

    return "ebeam_taper_te1550(" + ('%.3f-%.3f-%.3f' % (self.wg_width1,self.wg_width2,self.wg_length) ) + ")"



# $description: Double-bus ring resonator sweep (EBeam)
# $show-in-menu
# $group-name: Examples_EBeam
# $menu-path: siepic_menu.exlayout.begin
# Unit testing for all library layout fixed cells and PCells

'''
Scripted layout for ring resonators using SiEPIC-Tools
in the SiEPIC-EBeam-PDK "EBeam" technology

by Lukas Chrostowski, 2020-2023
'''

print('SiEPIC_EBeam_PDK: Example - Ring_resonator_sweep.py')
import pya
from pya import *

import SiEPIC
from SiEPIC._globals import Python_Env
from SiEPIC.scripts import zoom_out, export_layout
import os

if Python_Env == 'Script':
    try:
        # For external Python mode, when installed using pip install siepic_ebeam_pdk
        import siepic_ebeam_pdk
    except:
        # Load the PDK from a folder, e.g, GitHub, when running externally from the KLayout Application
        import os, sys
        path_GitHub = os.path.expanduser('~/Documents/GitHub/')
        sys.path.append(os.path.join(path_GitHub, 'SiEPIC_EBeam_PDK/klayout'))
        import siepic_ebeam_pdk

tech_name = 'EBeam'

# Example layout function
def dbl_bus_ring_res():

    # Import functions from SiEPIC-Tools
    from SiEPIC.extend import to_itype
    from SiEPIC.scripts import connect_cell, connect_pins_with_waveguide
    from SiEPIC.utils.layout import new_layout, floorplan

    # Create a layout for testing a double-bus ring resonator.
    # uses:
    #  - the SiEPIC EBeam Library
    # creates the layout in the presently selected cell
    # deletes everything first
    
    # Configure parameter sweep  
    pol = 'TE'
    if pol == 'TE':
        sweep_radius = [3,       5, 5, 5,          10, 10, 10, 10, 10]
        sweep_gap    = [0.07, 0.07, 0.08, 0.09, 0.07, 0.08, 0.09, 0.10, 0.11]
        x_offset = 67
        wg_bend_radius = 5
    else:
        sweep_radius = [30, 30, 30, 30, 30, 30, 30, 30]
        sweep_gap    = [0.25, 0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.60]
        x_offset = 67
        wg_bend_radius = 15

    wg_width = 0.5
    
    '''
    Create a new layout using the EBeam technology,
    with a top cell
    and Draw the floor plan
    '''    
    cell, ly = new_layout(tech_name, 'top', GUI=True, overwrite = True)
    floorplan(cell, 605e3, 410e3)

    if SiEPIC.__version__ < '0.5.1':
        pya.MessageBox.warning("Errors", "This example requires SiEPIC-Tools version 0.5.1 or greater.", pya.MessageBox.Ok)

    # Layer mapping:
    LayerSiN = ly.layer(ly.TECHNOLOGY['Si'])
    fpLayerN = cell.layout().layer(ly.TECHNOLOGY['FloorPlan'])
    TextLayerN = cell.layout().layer(ly.TECHNOLOGY['Text'])
    
    
    # Create a sub-cell for our Ring resonator layout
    top_cell = cell
    dbu = ly.dbu
    cell = cell.layout().create_cell("RingResonator")
    if pol == 'TE':
        t = Trans(Trans.R0, 40 / dbu, 12 / dbu)
    else:
        # rotate the layout since EBeam TM grating couplers have an angle that is negative
        t = Trans(Trans.R180, 560 / dbu, 393 / dbu)

    # place the cell in the top cell
    top_cell.insert(CellInstArray(cell.cell_index(), t))
    
    # Import cell from the SiEPIC EBeam Library
    cell_ebeam_gc = ly.create_cell("GC_%s_1550_8degOxide_BB" % pol, "EBeam")
    # get the length of the grating coupler from the cell
    gc_length = cell_ebeam_gc.bbox().width()*dbu
    # spacing of the fibre array to be used for testing
    GC_pitch = 127

    # Loop through the parameter sweep
    for i in range(len(sweep_gap)):
        
        # place layout at location:
        if i==0:
            x=0
        else:
            # next device is placed at the right-most element + length of the grating coupler
            x = inst_dc2.bbox().right*dbu + gc_length + 1
        
        # get the parameters
        r = sweep_radius[i]
        g = sweep_gap[i]
        
        # Grating couplers, Ports 0, 1, 2, 3 (from the bottom up)
        instGCs = []
        for i in range(0,4):
            t = Trans(Trans.R0, to_itype(x,dbu), i*127/dbu)
            instGCs.append( cell.insert(CellInstArray(cell_ebeam_gc.cell_index(), t)) )
        
        # Label for automated measurements, laser on Port 2, detectors on Ports 1, 3, 4
        t = Trans(Trans.R90, to_itype(x,dbu), to_itype(GC_pitch*2,dbu))
        text = Text ("opt_in_%s_1550_device_RingDouble%sr%sg%s" % (pol.upper(), pol.upper(),r,int(round(g*1000))), t)
        text.halign = 1
        cell.shapes(TextLayerN).insert(text).text_size = 5/dbu
        
        '''
        # Label for automated measurements, laser on Port 1, detectors on Ports 2, 3
        # this will cause an error in the verification since the 4th detector is not connected
        t = Trans(Trans.R0, to_itype(x,dbu), to_itype(GC_pitch*3,dbu))
        text = Text ("opt_in_%s_1550_device_RingDouble%sr%sg%sB" % (pol.upper(), pol.upper(),r,int(round(g*1000))), t)
        text.halign = 1
        cell.shapes(TextLayerN).insert(text).text_size = 5/dbu
        '''
          
        # Ring resonator from directional coupler PCells
        cell_dc = ly.create_cell("ebeam_dc_halfring_straight", "EBeam", { "r": r, "w": wg_width, "g": g, "bustype": 0 } )
        y_ring = GC_pitch*3/2
        # first directional coupler
        t1 = Trans(Trans.R270, to_itype(x+wg_bend_radius, dbu), to_itype(y_ring, dbu))
        inst_dc1 = cell.insert(CellInstArray(cell_dc.cell_index(), t1))
        # add 2nd directional coupler, snapped to the first one
        inst_dc2 = connect_cell(inst_dc1, 'pin2', cell_dc, 'pin4')
        
        # Create paths for waveguides, with the type defined in WAVEGUIDES.xml in the PDK
        waveguide_type='Strip TE 1550 nm, w=500 nm'
        
        # GC1 to bottom-left of ring pin3
        connect_pins_with_waveguide(instGCs[1], 'opt1', inst_dc1, 'pin3', waveguide_type=waveguide_type)
        
        # GC2 to top-left of ring pin1
        connect_pins_with_waveguide(instGCs[2], 'opt1', inst_dc1, 'pin1', waveguide_type=waveguide_type)
        
        # GC0 to top-right of ring
        connect_pins_with_waveguide(instGCs[0], 'opt1', inst_dc2, 'pin1', waveguide_type=waveguide_type)
        
        # GC3 to bottom-right of ring
        connect_pins_with_waveguide(instGCs[3], 'opt1', inst_dc2, 'pin3', waveguide_type=waveguide_type)

    # Introduce an error, to demonstrate the Functional Verification
    inst_dc2.transform(Trans(1000,-1000))


    return ly, cell
    
ly, cell = dbl_bus_ring_res()

# Zoom out
zoom_out(cell)

# Save
path = os.path.dirname(os.path.realpath(__file__))
filename = 'Test_structures_ring_resonators'
file_out = export_layout(cell, path, filename, relative_path = '', format='oas', screenshot=True)

from SiEPIC.verification import layout_check
print('SiEPIC_EBeam_PDK: example_Ring_resonator_sweep.py - verification')

file_lyrdb = os.path.join(path,filename+'.lyrdb')
num_errors = layout_check(cell = cell, verbose=False, GUI=True, file_rdb=file_lyrdb)

if Python_Env == 'Script':
    from SiEPIC.utils import klive
    klive.show(file_out, lyrdb_filename=file_lyrdb, technology=tech_name)

print('SiEPIC_EBeam_PDK: example_Ring_resonator_sweep.py - done')



import os, sys
import SiEPIC
from SiEPIC.utils import get_technology, get_technology_by_name
from pya import *
import math

path = os.path.dirname(os.path.abspath(__file__))

def linspace_without_numpy(low, high, length):
    step = ((high-low) * 1.0 / length)
    return [low+i*step for i in range(length)]

def pin(w,pin_text, trans, LayerPinRecN, dbu, cell):
  """
  w: Waveguide Width, e.g., 500 (in dbu)
  pin_text: Pin Text, e.g., "pin1"
  trans: e.g., trans =  Trans(0, False, 0, 0)  - first number is 0, 1, 2, or 3.
  pinrec: PinRec Layer, e.g., layout.layer(TECHNOLOGY['PinRec']))
  devrec: DevRec Layer, e.g., layout.layer(TECHNOLOGY['DevRec']))
  """
  
  # Create the pin, as short paths:
  from SiEPIC._globals import PIN_LENGTH

  pin = trans*Path([Point(-PIN_LENGTH/2, 0), Point(PIN_LENGTH/2, 0)], w)
  cell.shapes(LayerPinRecN).insert(pin)
  text = Text (pin_text, trans)
  shape = cell.shapes(LayerPinRecN).insert(text)
  shape.text_size = w*0.8

  # print("Done drawing the layout for - pin" )



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class ebeam_taper(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the strip waveguide taper.
  """

  def __init__(self):

    # Important: initialize the super class
    super(ebeam_taper, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # Button to launch separate window
    self.param("documentation", self.TypeCallback, "Open documentation in web browser")
#    self.param("simulation", self.TypeCallback, "Launch simulation GUI")
    
    # declare the parameters
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("wg_width1", self.TypeDouble, "Waveguide Width1", default = 0.5)
    self.param("wg_width2", self.TypeDouble, "Waveguide Width2", default = 3)
    self.param("wg_length", self.TypeDouble, "Waveguide Length", default = 10)
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    # hidden parameters, can be used to query this component:
    self.param("p1", self.TypeShape, "DPoint location of pin1", default = Point(-10000, 0), hidden = True, readonly = True)
    self.param("p2", self.TypeShape, "DPoint location of pin2", default = Point(0, 10000), hidden = True, readonly = True)

  def callback(self, layout, name, states):
    ''' Callback for PCell, to launch documentation viewer
        https://www.klayout.de/doc/code/class_PCellDeclaration.html#method9
    '''
    if name == 'documentation':
        url = 'https://github.com/SiEPIC/SiEPIC_EBeam_PDK/blob/master/Documentation/Taper/Summary_Taper.pdf'
        import webbrowser
        webbrowser.open_new(url)

    

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "ebeam_taper(R=" + ('%.3f-%.3f-%.3f' % (self.wg_width1,self.wg_width2,self.wg_length) ) + ")"

  def can_create_from_shape_impl(self):
    return False


  def produce(self, layout, layers, parameters, cell):
    """
    coerce parameters (make consistent)
    """
    self._layers = layers
    self.cell = cell
    self._param_values = parameters
    self.layout = layout
    shapes = self.cell.shapes


    # cell: layout cell to place the layout
    # LayerSiN: which layer to use
    # w: waveguide width
    # length units in dbu

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    
    LayerSi = self.silayer
    LayerSiN = self.silayer_layer
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    
    w1 = int(round(self.wg_width1/dbu))
    w2 = int(round(self.wg_width2/dbu))
    length = int(round(self.wg_length/dbu))

    pts = [Point(0,-w1/2), Point(0,w1/2), Point(length,w2/2), Point(length,-w2/2)]
    shapes(LayerSiN).insert(Polygon(pts))

    
    # Create the pins on the waveguides, as short paths:
    from SiEPIC._globals import PIN_LENGTH as pin_length
    
    # Pin on the left side:
    p1 = [Point(pin_length/2,0), Point(-pin_length/2,0)]
    p1c = Point(0,0)
    self.set_p1 = p1c
    self.p1 = p1c
    pin = Path(p1, w1)
    shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, 0, 0)
    text = Text ("opt1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Pin on the right side:
    p2 = [Point(length-pin_length/2,0), Point(length+pin_length/2,0)]
    p2c = Point(length, 0)
    self.set_p2 = p2c
    self.p2 = p2c
    pin = Path(p2, w2)
    shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, length, 0)
    text = Text ("opt2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    shape.text_halign = 2

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    path = Path([Point(0,0),Point(length,0)],w2+w1*2)
    shapes(LayerDevRecN).insert(path.simple_polygon())


    # Compact model information
    t = Trans(Trans.R0, w1/10, 0)
    text = Text ("Lumerical_INTERCONNECT_library=None", t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = length/100
    t = Trans(Trans.R0, length/10, w1/4)
    text = Text ('Component=ebeam_taper', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = length/100
    t = Trans(Trans.R0, length/10, w1/2)
    text = Text ('Spice_param:wg_width1=%.3fu wg_width2=%.3fu wg_length=%.3fu'% (self.wg_width1,self.wg_width2,self.wg_length), t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = length/100

    return "ebeam_taper(" + ('%.3f-%.3f-%.3f' % (self.wg_width1,self.wg_width2,self.wg_length) ) + ")"



# Python script

"""
This file is part of the SiEPIC_EBeam_PDK
by Lukas Chrostowski (c) 2015-2023

This Python file implements a library called "SiEPIC_EBeam", consisting of mature components that 
have Layouts and Compact Models for circuit simulations:
 - Fixed GDS cell components: imported from SiEPIC-EBeam.gds
 - PCells:
    - ebeam_dc_halfring_straight
    - ebeam_bragg_te1550: waveguide Bragg grating 
		- ebeam_taper_te1550: Waveguide Taper
		- Waveguide_bump
		- 
		- Waveguide_Bend 

NOTE: after changing the code, the macro needs to be rerun to install the new
implementation. The macro is also set to "auto run" to install the PCell 
when KLayout is run.

Crash warning:
 https://www.klayout.de/forum/discussion/734
 This library has nested PCells. Running this macro with a layout open may
 cause it to crash. Close the layout first before running.

Version history:

Lukas Chrostowski           2015/11/05 - 2015/11/10
 - Double-bus ring resonator
 - waveguide bends
 - PCell parameter functions
 - polygon text
 - PCell calling another PCell - TestStruct_DoubleBus_Ring

Lukas Chrostowski           2015/11/14
 - fix for rounding error in "DoubleBus_Ring"

Lukas Chrostowski           2015/11/15
 - fix for Python 3.4: print("xxx")
 
Lukas Chrostowski           2015/11/17
 - update "layout_waveguide_rel" to use the calculated points_per_circle(radius)

Lukas Chrostowski           2015/11/xx
 - Waveguide based on bends, straight waveguide.

Lukas Chrostowski           2015/12/3
 - Bragg grating

Lukas Chrostowski           2016/01/17
 - Taper, matching EBeam CML component
 
Lukas Chrostowski           2016/01/20
 - (sinusoidal) Bragg grating
 
Lukas Chrostowski           2016/05/27
 - SWG_waveguide
 - SWG_to_strip_waveguide

S. Preble                   2016/08/26
 - Double Bus Ring Pin's shifted - text now is in the middle of the pin path

Lukas Chrostowski           2016/11/06
 - waveguide bump, to provide a tiny path length increase

Lukas Chrostowski           2017/02/14
 - renaming "SiEPIC" PCells library to "SiEPIC-EBeam PCells", update for Waveguide_Route
 - code simplifications: Box -> Box

Lukas Chrostowski           2017/03/08
 - S-Bend - 

Lukas Chrostowski           2017/03/18
 - ebeam_dc_halfring_straight, with TE/TM support.  Zeqin Lu adding CML for this component. 
 
Lukas Chrostowski 2017/12/16
 - compatibility with KLayout 0.25 and SiEPIC-Tools

Lukas Chrostowski 2020/02/22
 - contra directional coupler; moved from Dev library; now has a compact model generator
  
todo:
replace:     
 layout_arc_wg_dbu(self.cell, Layerm1N, x0,y0, r_m1_in, w_m1_in, angle_min_doping, angle_max_doping)
with:
 self.cell.shapes(Layerm1N).insert(Polygon(arc(w_m1_in, angle_min_doping, angle_max_doping) + 
 [Point(0, 0)]).transformed(t))


Lukas Chrostowski 2020/06/21
 - Waveguide update, cellName

Mustafa Hammood 2020/06/26
 - major refactoring and splitting of individual classes into sub files

Lukas 2021/04/01
 - fixing loading library (previous collisions with other PDKs)

Lukas 2023/11
 - compatibility with PyPI usage of KLayout
 
"""

verbose=False

import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
import os
import pathlib
import sys


from SiEPIC._globals import KLAYOUT_VERSION, KLAYOUT_VERSION_3
if KLAYOUT_VERSION < 28:
    question = pya.QMessageBox()
    question.setStandardButtons(pya.QMessageBox.Ok)
    question.setText("SiEPIC-EBeam is not compatible with older versions (<0.28) of KLayout.")
    KLayout_link0='https://www.klayout.de/build.html'
    question.setInformativeText("\nSiEPIC-EBeam is not compatible with older versions (<0.28) of KLayout.\nPlease download an install the latest version, from %s" % (KLayout_link0))
    pya.QMessageBox_StandardButton(question.exec_())


dir_path = os.path.dirname(os.path.realpath(__file__))
if dir_path not in sys.path:
    sys.path.append(dir_path)

files = [f for f in os.listdir(os.path.join(os.path.dirname(
    os.path.realpath(__file__)),'pcells_EBeam')) if '.py' in pathlib.Path(f).suffixes  and '__init__' not in f]
import pcells_EBeam ### folder name ###
import importlib
importlib.invalidate_caches()
pcells_=[]
for f in files:
    module = 'pcells_EBeam.%s' % f.replace('.py','')  ### folder name ###
    if verbose:
        print(' - found module: %s' % module)
    m = importlib.import_module(module) 
    if verbose:
        print(m)
    pcells_.append(importlib.reload(m))


class SiEPIC_EBeam_Library(Library):
  """
  The library where we will put the PCells and GDS into 
  """

  def __init__(self):
  
    tech_name = 'EBeam'
    library = tech_name
    self.technology=tech_name

    if verbose:
        print("Initializing '%s' Library." % library)

    # Set the description
    self.description = "v0.4.3, Components with models"

    # Save the path, used for loading WAVEGUIDES.XML
    import os
    self.path = os.path.dirname(os.path.realpath(__file__))

    # Import all the GDS files from the tech folder
    import os, fnmatch
    dir_path = os.path.normpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../gds/EBeam"))
    if verbose:
        print('  library path: %s' % dir_path)
    search_str = '*.[Oo][Aa][Ss]' # OAS
    for root, dirnames, filenames in os.walk(dir_path, followlinks=True):
        for filename in fnmatch.filter(filenames, search_str):
            file1=os.path.join(root, filename)
            if verbose:
                print(" - reading %s" % file1 )
            self.layout().read(file1)
    search_str = '*.[Gg][Dd][Ss]' # GDS
    for root, dirnames, filenames in os.walk(dir_path, followlinks=True):
        for filename in fnmatch.filter(filenames, search_str):
            file1=os.path.join(root, filename)
            if verbose:
                print(" - reading %s" % file1 )
            self.layout().read(file1)
                        
    # Create the PCell declarations
    for m in pcells_:
        mm = m.__name__.replace('pcells_EBeam.','')
        mm2 = m.__name__+'.'+mm+'()'
        if verbose:
            print(' - register_pcell %s, %s' % (mm,mm2))
        self.layout().register_pcell(mm, eval(mm2))
                
    if verbose:
        print(' done with pcells')
    
    # Register us the library with the technology name
    # If a library with that name already existed, it will be replaced then.
    self.register(library)
 
# Instantiate and register the library
SiEPIC_EBeam_Library()



""" OID: orthogonal ID for polarization, 1=TE (first mode)
"""

from copy import deepcopy
from pathlib import Path
import numpy as np
from SiEPIC.opics.components import componentModel
from SiEPIC.opics.utils import LUT_reader
from numpy import complex128, ndarray
from pathlib import PosixPath

datadir = Path(str(Path(__file__).parent.parent.parent)) / "CML/EBeam/source_data"



class contra_directional_coupler(componentModel):
    """
    Bragg Grating-assisted Contra Directional Coupler. 
    This component operates as an optical add-drop multiplexer (AODM) or filter.

    Model schematic:
    ~~~~~~~~~~~~~~~

    0 ┌───┐             ┌───┐ 2
      └───┼──┐       ┌──┼───┘
          └──┼───────┼──┘
             │┼┼┼┼┼┼┼│
          ┌──┼───────┼──┐
      ┌───┼──┘       └──┼───┐
    1 └───┘             └───┘ 3
    """

    cls_attrs = {"wg1_width": 0, "wg2_width": 0, "corrugation1_width": 0, "corrugation2_width": 0, "gap": 0, "grating_period": 0, "number_of_periods": 0, "sinusoidal": 0, "apodization_index": 0, "rib": 0}
    valid_OID = [1]
    ports = 4


    def __init__(
        self,
        f: ndarray = None,
        wg1_width: float = 560e-9,
        wg2_width: float = 440e-9,
        corrugation1_width: float = 50e-9,
        corrugation2_width: float = 25e-9,
        gap: float = 100e-9,
        grating_period: float = 316e-9,
        number_of_periods: int = 1000,
        sinusoidal: bool = False,
        apodization_index: float = 10.0,
        rib: bool = False,
        OID: int = 1,
    ) -> None:
        data_folder = datadir / "contraDC"
        filename = "contraDC.xml"
        '''
        file_search =  "w1=" + "%.0f"%(wg1_width*1e9) + ",w2=" + "%.0f"%(wg2_width*1e9) + ",dW1=" + "%.0f"%(corrugation1_width*1e9) + ",dW2=" + "%.0f"%(corrugation2_width*1e9) +  ",gap=" + "%.0f"%(gap*1e9) + ",p=" + "%.1f"%(grating_period*1e9) + ",N=" + "%.0f"%(number_of_periods) + ",s=" + str(1 if sinusoidal else 0) +  ",a=" + "%.2f"%apodization_index +  ",rib=" + str(1 if rib else 0) + ",pol=" + str(OID-1)
        import os, fnmatch
        files = fnmatch.filter(os.listdir(data_folder), file_search + '*')
        if not files:
            raise Exception ('The contra-directional coupler Compact Model Library does not have data for the chosen parameters. Requested file: ' + file_search);
        else:
            filename = files[0] # pick the first one. Could be improved to pick the one with the highest resolution, etc., or give people the choice.
            print('contra directional coupler, data file: %s' % filename)    
        '''
        LUT_attrs_ = deepcopy(self.cls_attrs)
        LUT_attrs_["wg1_width"] = wg1_width,
        LUT_attrs_["wg2_width"] = wg2_width,
        LUT_attrs_["corrugation1_width"] = corrugation1_width,
        LUT_attrs_["corrugation2_width"] = corrugation2_width,
        LUT_attrs_["gap"] = gap,
        LUT_attrs_["grating_period"] = grating_period,
        LUT_attrs_["number_of_periods"] = number_of_periods,
        LUT_attrs_["sinusoidal" ] = sinusoidal,
        LUT_attrs_["apodization_index" ] = apodization_index,
        LUT_attrs_["rib"] = rib,
        LUT_attrs_["OID"] =  OID,

        super().__init__(
            f=f,
            data_folder=data_folder,
            filename=filename,
            nports=4,
            sparam_attr="contra_directional_coupler",
            **LUT_attrs_
        )

        if OID in self.valid_OID:
            self.s = self.load_sparameters(data_folder, filename)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))
        self.component_id = "contra_directional_coupler"



class ebeam_bdc_te1550(componentModel):
    """
    50/50% broadband directional 3-dB couplers. Two 3-dB couplers can be used to make an unbalanced Mach-Zehnder Interferometer (MZI),
    showing a large extinction ratio. The advantage of this device compared to the Y-Branch is that it has 2x2 ports,
    thus the MZI has two outputs. Compared to the directional coupler, it is less wavelength sensitive.

    Model schematic:
    ~~~~~~~~~~~~~~~

    0 ┌───┐             ┌───┐ 2
      └───┼──┐       ┌──┼───┘
          └──┼───────┼──┘
             │┼┼┼┼┼┼┼│
          ┌──┼───────┼──┐
      ┌───┼──┘       └──┼───┐
    1 └───┘             └───┘ 3
    """

    cls_attrs = {"height": 0, "width": 0}
    valid_OID = [1]
    ports = 4

    def __init__(
        self,
        f: ndarray = None,
        height: float = 220e-9,
        width: float = 500e-9,
        OID: int = 1,
    ) -> None:
        data_folder = datadir / "bdc_TE_source"
        filename = "bdc_lookup_table.xml"

        LUT_attrs_ = deepcopy(self.cls_attrs)
        LUT_attrs_["height"] = height
        LUT_attrs_["width"] = width
        super().__init__(
            f=f,
            data_folder=data_folder,
            filename=filename,
            nports=4,
            sparam_attr="bdc_sparam",
            **LUT_attrs_
        )
        if OID in self.valid_OID:
            self.s = self.load_sparameters(data_folder, filename)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))
        self.component_id = "ebeam_bdc_te1550"


class DC_temp(componentModel):
    """
    The directional coupler is commonly used for splitting and combining light in photonics.
    It consists of two parallel waveguides where the coupling coefficient is influenced by the
    waveguide length and the distance between waveguides.

    Model schematic:
    ~~~~~~~~~~~~~~~

    1                      3
     ##                 ##
      ###             ###
        ###         ###
          ###     ###
              ###

    0 ###################  2
    """

    cls_attrs = {"Lc": 0}
    valid_OID = [1]
    ports = 4

    def __init__(self, f=None, Lc=0, OID=1):
        data_folder = datadir / "ebeam_dc_te1550"
        filename = "dc_map.xml"

        LUT_attrs_ = deepcopy(self.cls_attrs)
        LUT_attrs_["Lc"] = Lc
        super().__init__(
            f=f,
            data_folder=data_folder,
            filename=filename,
            nports=4,
            sparam_attr="s-param",
            **LUT_attrs_
        )
        if OID in self.valid_OID:
            self.s = self.load_sparameters(data_folder, filename)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))
        self.component_id = "Ebeam_DC"


class DC_halfring(componentModel):
    """
    Models evanescent coupling region between a straight waveguide and a bent radius of length pi*radius um. Useful for filters, sensors.

    Model schematic:
    ~~~~~~~~~~~~~~~

    1                      3
     ##                 ##
      ###             ###
        ###         ###
          ###     ###
              ###

    0 ###################  2
    """

    cls_attrs = {
        "CoupleLength": 0,
        "gap": 100e-9,
        "radius": 5e-6,
        "thickness": 220e-9,
        "width": 500e-9,
    }
    valid_OID = [1]
    ports = 4

    def __init__(
        self,
        f: ndarray = None,
        CoupleLength: int = 0,
        gap: float = 100e-9,
        radius: float = 5e-6,
        thickness: float = 220e-9,
        width: float = 500e-9,
        OID: int = 1,
    ) -> None:
        data_folder = datadir / "ebeam_dc_halfring_straight"
        filename = "te_ebeam_dc_halfring_straight.xml"

        LUT_attrs_ = deepcopy(self.cls_attrs)
        LUT_attrs_["CoupleLength"] = CoupleLength
        LUT_attrs_["gap"] = gap
        LUT_attrs_["radius"] = radius
        LUT_attrs_["thickness"] = thickness
        LUT_attrs_["width"] = width

        super().__init__(f=f, data_folder=data_folder, filename=filename,
                         nports=4, sparam_attr="s-param", **LUT_attrs_)
        if OID in self.valid_OID:
            self.s = self.load_sparameters(data_folder, filename)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))
        self.component_id = "Ebeam_DC_halfring"


class GC(componentModel):
    """
    Fully-etched fibre-waveguide grating couplers with sub-wavelength gratings showing high coupling efficiency as well as low
    back reflections for both transverse electric (TE) and transverse magnetic (TM) modes. EBeam fabrication cost is reduced
    by ~2-3X when eliminating the shallow etch.

    Model schematic:
    ~~~~~~~~~~~~~~~~
                     |
        ◄──────    │ │
               │ │ │ │
       ┌───────┤ │ │ │
    1  └───────┤ │ │ │  0
               │ │ │ │
                   │ │
                     |
    """

    cls_attrs = {"deltaw": 0, "height": 2.2e-07}
    valid_OID = [1]
    ports = 2

    def __init__(
        self, f: ndarray = None, deltaw: int = 0, height: float = 2.2e-07, OID: int = 1
    ) -> None:

        data_folder = datadir / "gc_source"
        filename = "GC_TE_lookup_table.xml"

        LUT_attrs_ = deepcopy(self.cls_attrs)
        LUT_attrs_["deltaw"] = deltaw
        LUT_attrs_["height"] = height
        super().__init__(
            f=f,
            nports=2,
            data_folder=data_folder,
            filename=filename,
            sparam_attr="gc_sparam",
            **LUT_attrs_
        )
        if OID in self.valid_OID:
            self.s = self.load_sparameters(data_folder, filename)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))

        self.component_id = "Ebeam_GC"


class Multimode(componentModel):
    valid_OID = [1, 2]
    ports = 2

    def __init__(self, f=None, OID=1):
        super().__init__(f, "", "")
        if OID in self.valid_OID and OID == 1:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))
            self.s[1, 0] = self.s[0, 1] = -2 * np.ones((self.f.shape[0]))
        elif OID in self.valid_OID and OID == 2:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))
            self.s[1, 0] = self.s[0, 1] = -5 * np.ones((self.f.shape[0]))
        self.component_id = "Ebeam_multimode"


class Terminator(componentModel):
    """
    This component is used to terminate a waveguide. This terminator is a nano-taper that spreads
    the light into the oxide and is used for efficient edge coupling. Even if a waveguide crosses near
    this taper end, the reflection is minimal. This is included in this model, 1 µm away, therefore,
    the model is a worst-case reflection. To terminate unused ports on components to avoid reflections,
    refer to Disconnected Waveguides.

    Model schematic:
    ~~~~~~~~~~~~~~~


      ┌┬──┐
    0 ││  ├────────┐
      ││  ├────────┘
      └┴──┘



    """

    valid_OID = [1]
    ports = 2

    def __init__(self, f=None, OID=1):
        data_folder = datadir / "ebeam_terminator_te1550"
        filename = "ebeam_terminator_te1550.npz"
        super().__init__(f=f, data_folder=data_folder, filename=filename)
        if OID in self.valid_OID:
            self.s = self.load_sparameters(data_folder=data_folder, filename=filename)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))
        self.component_id = "Ebeam_Terminator"


class TunableWG(componentModel):
    """
    Waveguides are components that guide waves. Although these are individual components that can
    be adjusted for use, it is recommended to draw paths in KLayout and convert them to waveguides
    using the built-in SiEPIC features.

    The behavior of tunable waveguides can be adjusted by modifying the `power` parameter.

    Model schematic:
    ~~~~~~~~~~~~~~~

    0 ┌─────────┐ 1
      └─────────┘

    """

    cls_attrs = {"power": 0}
    valid_OID = [1, 2]
    ports = 2

    def __init__(
        self,
        f: ndarray = None,
        length: float = 5e-6,
        power: float = 0e-3,
        loss: int = 700,
        OID: int = 1,
    ) -> None:

        data_folder = datadir / "tunable_wg"
        filename = "wg_strip_tunable.xml"
        LUT_attrs_ = deepcopy(self.cls_attrs)
        LUT_attrs_["power"] = power

        if OID in self.valid_OID:
            self.s = self.load_sparameters(
                length=length, data_folder=data_folder, filename=filename, neff=None, ng=None, loss=loss)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))
        self.component_id = "Ebeam_TunableWG"

    def load_sparameters(
        self, length: float, data_folder: PosixPath, filename: str, neff: float, ng: float, loss: int
    ) -> ndarray:
        """overrides read s_parameters"""

        sfilename, _, _ = LUT_reader(data_folder, filename, self.componentParameters)
        self.sparam_file = sfilename[-1]
        filepath = data_folder / sfilename[-1]

        # Read info from waveguide s-param file
        with open(filepath, "r") as f:
            coeffs = f.readline().split()

        # Initialize array to hold s-params
        temp_s_ = np.zeros((len(self.f), self.nports, self.nports), dtype=complex128)

        alpha = loss / (20 * np.log10(np.exp(1)))

        # calculate angular frequency from frequency
        w = np.asarray(self.f) * 2 * np.pi

        # calculate center wavelength, effective index, group index, group dispersion
        lam0, ne, ng_, nd = (
            float(coeffs[0]),
            float(coeffs[1]),
            float(coeffs[3]),
            float(coeffs[5]),
        )

        if(neff):
            ne = neff
        if(ng):
            ng = ng
        else:
            ng = ng_

        # calculate angular center frequency
        w0 = (2 * np.pi * self.C) / lam0

        # calculation of K
        K = (
            2 * np.pi * ne / lam0
            + (ng / self.C) * (w - w0)
            - (nd * lam0**2 / (4 * np.pi * self.C)) * ((w - w0) ** 2)
        )

        # compute s-matrix from K and waveguide length
        temp_s_[:, 0, 1] = temp_s_[:, 1, 0] = np.exp(
            -alpha * length + (K * length * 1j)
        )

        s = temp_s_
        self.ng_ = ng
        self.alpha_ = alpha
        self.ne_ = ne
        self.nd_ = nd

        return s


class Waveguide(componentModel):
    """
    Waveguides are components that guide waves. Although these are individual components that can
    be adjusted for use, it is recommended to draw paths in KLayout and convert them to waveguides
    using the built-in SiEPIC features.

    Model schematic:
    ~~~~~~~~~~~~~~~~

    0 ┌─────────┐ 1
      └─────────┘

    """

    cls_attrs = {"length": 0e-6, "height": 220e-9, "width": 500e-9}
    valid_OID = [1, 2]
    ports = 2

    def __init__(
        self,
        f: ndarray = None,
        length: float = 5e-6,
        height: float = 220e-9,
        width: float = 500e-9,
        loss: int = 700,
        OID: int = 1,
    ) -> None:
        data_folder = datadir / "wg_integral_source"
        filename = "wg_strip_lookup_table.xml"

        LUT_attrs_ = deepcopy(self.cls_attrs)
        LUT_attrs_["height"] = height
        LUT_attrs_["width"] = width

        # length is not part of LUT attributes
        del LUT_attrs_["length"]

        super().__init__(
            f,
            length=length,
            data_folder=data_folder,
            filename=filename,
            loss=loss,
            **LUT_attrs_
        )

        if OID in self.valid_OID:
            self.s = self.load_sparameters(
                length=length, data_folder=data_folder, filename=filename, neff=None, ng=None, loss=loss)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))

        self.component_id = "Ebeam_WG"

    def load_sparameters(
        self, length: float, data_folder: PosixPath, filename: str, neff: float, ng: float, loss: int
    ) -> ndarray:
        """overrides read s_parameters"""

        sfilename, _, _ = LUT_reader(data_folder, filename, self.componentParameters)
        self.sparam_file = sfilename[-1]
        filepath = data_folder / sfilename[-1]

        # Read info from waveguide s-param file
        with open(filepath, "r") as f:
            coeffs = f.readline().split()

        # Initialize array to hold s-params
        temp_s_ = np.zeros((len(self.f), self.nports, self.nports), dtype=complex128)

        alpha = loss / (20 * np.log10(np.exp(1)))

        # calculate angular frequency from frequency
        w = np.asarray(self.f) * 2 * np.pi

        # calculate center wavelength, effective index, group index, group dispersion
        lam0, ne, ng_, nd = (
            float(coeffs[0]),
            float(coeffs[1]),
            float(coeffs[3]),
            float(coeffs[5]),
        )

        if(neff):
            ne = neff
        if(ng):
            ng = ng
        else:
            ng = ng_

        # calculate angular center frequency
        w0 = (2 * np.pi * self.C) / lam0

        # calculation of K
        K = (
            2 * np.pi * ne / lam0
            + (ng / self.C) * (w - w0)
            - (nd * lam0**2 / (4 * np.pi * self.C)) * ((w - w0) ** 2)
        )

        # compute s-matrix from K and waveguide length
        temp_s_[:, 0, 1] = temp_s_[:, 1, 0] = np.exp(
            -alpha * length + (K * length * 1j)
        )

        s = temp_s_
        self.ng_ = ng
        self.alpha_ = alpha
        self.ne_ = ne
        self.nd_ = nd

        return s


class Y(componentModel):
    r"""
    50/50 3dB splitter. Useful for splitting light, Mach-Zehner Interferometers, etc.
    The layout parameters for the device were taken from the journal paper below, and implemented in EBeam lithography.

    Model schematic:
    ~~~~~~~~~~~~~~~~

              ┌─────────┐ 1
              ├─┼───────┘
              │ │
    0 ┌───────┼─┤
      └───────┼─┤
              │ │
              ├─┼───────┐ 2
              └─────────┘

    """

    cls_attrs = {"height": 220e-9, "width": 500e-9}
    valid_OID = [1]
    ports = 3

    def __init__(
        self,
        f: ndarray = None,
        height: float = 220e-9,
        width: float = 500e-9,
        OID: int = 1,
    ) -> None:

        data_folder = datadir / "y_branch_source"
        filename = "y_lookup_table.xml"
        LUT_attrs_ = deepcopy(self.cls_attrs)
        LUT_attrs_["height"] = height
        LUT_attrs_["width"] = width

        # print(LUT_attrs_)
        super().__init__(
            f=f,
            nports=3,
            data_folder=data_folder,
            filename=filename,
            sparam_attr="y_sparam",
            **LUT_attrs_
        )
        if OID in self.valid_OID:
            self.s = self.load_sparameters(data_folder, filename)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))
        self.component_id = "Ebeam_Y"


class Switch(componentModel):
    """2x2 tunable optical switch component. Useful for switching the input optical power between two output ports."""

    cls_attrs = {"power": 0}
    valid_OID = [1]
    ports = 4

    def __init__(self, f: ndarray = None, power: float = 0e-3, OID: int = 1) -> None:
        data_folder = datadir / "2x2_switch"
        filename = "2x2_switch.xml"

        LUT_attrs_ = deepcopy(self.cls_attrs)
        LUT_attrs_["power"] = power
        super().__init__(f, data_folder, filename, 4, "switch_sparam", **LUT_attrs_)
        if OID in self.valid_OID:
            self.s = self.load_sparameters(data_folder, filename)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))
        self.component_id = "Ebeam_switch"


class ebeam_wg_integral_1550(componentModel):
    """
    Waveguides are components that guide waves. Although these are individual components that can
    be adjusted for use, it is recommended to draw paths in KLayout and convert them to waveguides
    using the built-in SiEPIC features.

    Model schematic:
    ~~~~~~~~~~~~~~~~

    0 ┌─────────┐ 1
      └─────────┘

    """

    cls_attrs = {"wg_length": 0e-6, "wg_height": 220e-9, "wg_width": 500e-9}
    valid_OID = [1, 2]
    ports = 2

    def __init__(
        self,
        f: ndarray = None,
        wg_length: float = 5e-6,
        wg_height: float = 220e-9,
        wg_width: float = 500e-9,
        loss: int = 700,
        OID: int = 1,
    ) -> None:
        data_folder = datadir / "wg_integral_source"
        filename = "wg_strip_lookup_table.xml"

        LUT_attrs_ = deepcopy(self.cls_attrs)
        LUT_attrs_["wg_height"] = wg_height
        LUT_attrs_["wg_width"] = wg_width

        # wg_length is not part of LUT attributes
        del LUT_attrs_["wg_length"]

        super().__init__(
            f,
            length=wg_length,
            data_folder=data_folder,
            filename=filename,
            loss=loss,
            **LUT_attrs_
        )

        if OID in self.valid_OID:
            self.s = self.load_sparameters(
                length=wg_length, data_folder=data_folder, filename=filename, neff=None, ng=None, loss=loss)
        else:
            self.s = np.zeros((self.f.shape[0], self.ports, self.ports))

        self.component_id = "Ebeam_WG"

    def load_sparameters(
        self, length: float, data_folder: PosixPath, filename: str, neff: float, ng: float, loss: int
    ) -> ndarray:
        """overrides read s_parameters"""

        sfilename, _, _ = LUT_reader(data_folder, filename, self.componentParameters)
        self.sparam_file = sfilename[-1]
        filepath = data_folder / sfilename[-1]

        # Read info from waveguide s-param file
        with open(filepath, "r") as f:
            coeffs = f.readline().split()

        # Initialize array to hold s-params
        temp_s_ = np.zeros((len(self.f), self.nports, self.nports), dtype=complex128)

        alpha = loss / (20 * np.log10(np.exp(1)))

        # calculate angular frequency from frequency
        w = np.asarray(self.f) * 2 * np.pi

        # calculate center wavelength, effective index, group index, group dispersion
        lam0, ne, ng_, nd = (
            float(coeffs[0]),
            float(coeffs[1]),
            float(coeffs[3]),
            float(coeffs[5]),
        )

        if(neff):
            ne = neff
        if(ng):
            ng = ng
        else:
            ng = ng_

        # calculate angular center frequency
        w0 = (2 * np.pi * self.C) / lam0

        # calculation of K
        K = (
            2 * np.pi * ne / lam0
            + (ng / self.C) * (w - w0)
            - (nd * lam0**2 / (4 * np.pi * self.C)) * ((w - w0) ** 2)
        )

        # compute s-matrix from K and waveguide length
        temp_s_[:, 0, 1] = temp_s_[:, 1, 0] = np.exp(
            -alpha * length + (K * length * 1j)
        )

        s = temp_s_
        self.ng_ = ng
        self.alpha_ = alpha
        self.ne_ = ne
        self.nd_ = nd

        return s


component_factory = dict(
    contra_directional_coupler = contra_directional_coupler,
    BDC = ebeam_bdc_te1550,
    ebeam_bdc_te1550 = ebeam_bdc_te1550,
    DC_halfring=DC_halfring,
    # DC_temp=DC_temp,
    GC=GC,
    # Multimode=Multimode,
    # Path=Path,
    Switch=Switch,
    # Terminator=Terminator,
    TunableWG=TunableWG,
    Waveguide=Waveguide,
    Y=Y,
    ebeam_y_1550=Y,
    ebeam_gc_te1550=GC,
    GC_TE_1550_8degOxide_BB = GC,
    ebeam_wg_integral_1550=ebeam_wg_integral_1550,
)

components_list = list(component_factory.keys())
__all__ = components_list
__version__ = "0.4.4"

if __name__ == "__main__":
    import SiEPIC.opics as op

    w = np.linspace(1.52, 1.58, 3) * 1e-6
    f = op.C / w
    c = ebeam_bdc_te1550(f=f)
    s = c.get_data()

    print(components_list)
    print(datadir)



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class Waveguide_SWG_to_Strip(pya.PCellDeclarationHelper):
  """
  Input: length, period_strip, period_swg, wg_width_strip, wg_width_swg, duty_strip, duty_swg
  continuously (linearly) variable period, width, duty along the length
  
  Updated by Lukas Chrostowski, Leanne Dias, Connor Mosquera 2020/05
  """

  def __init__(self):

    # Important: initialize the super class
    super(Waveguide_SWG_to_Strip, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("wavelength", self.TypeDouble, "Wavelength", default = 1310)
    self.param("fishbone", self.TypeBoolean, "Fishbone", default = False)     
    self.param("length", self.TypeDouble, "Waveguide length", default = 10.0)     
    self.param("taper_fraction", self.TypeDouble, "Strip taper length fraction (0 to 1)", default = 1)     
    self.param("period_strip", self.TypeDouble, "SWG Period at strip end (microns)", default = 0.200)     
    self.param("period_swg", self.TypeDouble, "SWG Period at SWG end (microns)", default = 0.200)     
    self.param("wg_width_strip", self.TypeDouble, "Strip Waveguide width at the taper end (microns)", default = 0.5)     
#    self.param("wg_width_swg_taperend", self.TypeDouble, "Waveguide width at SWG taper end (microns)", default = 0.4)     
    self.param("wg_width_swg", self.TypeDouble, "SWG Waveguide at end (microns)", default = 0.5)     
    self.param("wg_width_taper", self.TypeDouble, "Nanotaper width at taper end (microns)", default = 0.06)     
    self.param("duty_strip", self.TypeDouble, "SWG duty cycle at strip end (0 to 1)", default = 0.500)     
    self.param("duty_swg", self.TypeDouble, "SWG duty cycle at SWG end (0 to 1)", default = 0.700)     
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Waveguide_SWG_to_Strip_%s-%.3f-%.3f-%.3f" % \
    (self.length, self.period_swg, self.wg_width_swg, self.duty_swg)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    debug = False
    # Layout and layers:
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    LayerSiSPN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

  
    # fetch the PCell parameters, and calculate a few things
    from SiEPIC.extend import to_itype
    wg_width_swg = to_itype(self.wg_width_swg,dbu)
    wg_width_strip = to_itype(self.wg_width_strip,dbu)
    length = to_itype(self.length, dbu)
    w1 = to_itype(self.wg_width_taper,dbu)
    w2 = to_itype(self.wg_width_strip,dbu)

    # Determine the period such that the waveguide length is as desired.  Slight adjustment to period
    N_boxes = int(round(self.length / (self.period_swg+self.period_strip)*2.0)-0.5)
    grating_period = self.length / (N_boxes) / dbu
    if debug:
      print("N boxes: %s, grating_period: %s" % (N_boxes, grating_period) )

    # taper length, minus last SWG block
    # taper_length = int(round(self.taper_fraction * self.length / dbu))
    taper_length = (N_boxes-1) * grating_period

    # Find out what is the delta width required between 
    # a) the SWG end and b) the SWG with nanotaper
    # Done by simulating the neff versus SWG waveguide width, with an without a 60 nm
    w = wg_width_swg
    if self.fishbone:
        dw = 0
    elif self.wavelength == 1310:
        dw = -3.3736e-5*w*w + 0.0618752*w  + 3.88577
    else:
        dw = 0
    wg_width_swg_taperend = wg_width_swg - dw

    
    # Draw the SWG waveguide:
    
    x = - self.period_swg * self.duty_swg / 2 / dbu
    for i in range(0,N_boxes):
      local_duty = 1.0 * (N_boxes - i) / N_boxes * self.duty_swg + 1.0*i / N_boxes * self.duty_strip 
      local_period = to_itype((1.0*(N_boxes - i) / N_boxes * self.period_swg + 1.0*i / N_boxes * self.period_strip ),dbu)
      local_wg_width = (1.0*(N_boxes - i) / N_boxes * wg_width_swg_taperend + 1.0*i / N_boxes * wg_width_strip  )
      if (i==0) | (i==N_boxes):
        if debug:
          print("local_duty: %s, local_period: %s, local_wg_width: %s" % (local_duty, local_period, local_wg_width) )     
      local_box_width = int(round(local_period*local_duty))
#      x = int(round((i * local_period - local_box_width/2)))
      if i != 0:
        box1 = Box(x, -local_wg_width/2, x + local_box_width, local_wg_width/2)
      else:
      # Last SWG, that is a bit larger than the one connected to the taper  
        box1 = Box(x, -wg_width_swg/2, x + local_box_width, wg_width_swg/2)
      shapes(LayerSiN).insert(box1)
      x = x + int(round((local_period)))
      

    # Taper
    if self.fishbone:
        taper_length = length
    pts = [Point(length-taper_length,-w1/2), Point(length-taper_length,w1/2), Point(length,w2/2), Point(length,-w2/2)]
    shapes(LayerSiN).insert(Polygon(pts))


    # Pins on the waveguide:
    
    from SiEPIC._globals import PIN_LENGTH as pin_length
    
    w = to_itype(self.wg_width_swg, dbu)
    t = Trans(Trans.R0, 0,0)
    pin = Path([Point(pin_length/2, 0), Point(-pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    w = to_itype(max ( self.wg_width_strip, self.wg_width_taper), dbu)
    t = Trans(Trans.R0, length,0)
    pin = Path([Point(-pin_length/2, 0), Point(pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    shape.text_halign = 2

    '''
    # Compact model information
    t = Trans(Trans.R0, 0, 0)
    text = Text ('Lumerical_INTERCONNECT_library=Design kits/ebeam_v1.2', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, length/10, 0)
    text = Text ('Component=NO_MODEL_AVAILABLE', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, length/9, -w*2)
    text = Text \
      ('Spice_param:length=%.3fu period_swg=%.3fu period_strip=%.3fu wg_width_swg=%.3fu wg_width_strip=%.3fu duty_swg=%.3f duty_strip=%.3f ' %\
      (self.length, self.period_swg, (self.period_strip), self.wg_width_swg, self.wg_width_strip, self.duty_swg, self.duty_strip), t )
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    blabla
    '''
    
    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    points = [pya.Point(0,0), pya.Point(length, 0)]
    path = pya.Path(points,w*5)
    shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class Bezier_Bend(pya.PCellDeclarationHelper):
  """
  Input: bezier number, effective radius, waveguide width, layers
  draws a 90 degree bezier bend with a given bezier factor
  Usage: instantiate, and use transformations (rotation)
  
  This implementation is Python translation of the 90 degree Bezier bend ample script for Mentor Graphics Pyxis
    by Jonas Flueckiger and Lukas Chrostowski's
    influenced by www.tinaja.com/glib/cubemath.pdf implementation
    
  Author: Mustafa Hammood     mustafa@ece.ubc.ca
  """

  def __init__(self):

    # Important: initialize the super class
    super(Bezier_Bend, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("bezier_num", self.TypeDouble, "Bezier factor", default = 25)     
    self.param("eff_r", self.TypeDouble, "Effective bend radius (microns)", default = 5)
    self.param("wg_width", self.TypeDouble, "Waveguide width (microns)", default = 0.500)     
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
#    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Bezier_Number_%.3f-%.3f" % (self.bezier_num/1000, self.wg_width/1000)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
  
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

    
    

    from SiEPIC.utils import arc_bezier
    from SiEPIC._globals import PIN_LENGTH as pin_length
    from SiEPIC.extend import to_itype
    wg_width = self.wg_width
    w = self.wg_width/dbu

    r = self.eff_r/dbu
    bezier = self.bezier_num/100
    
    x = -r
    y = r
    
    points = arc_bezier(r, 0, 360, bezier)
    a = pya.Path(points, w).simple_polygon()
    d= [each for each in a.each_point()]
    pt_idx = int(len(d)/2)
    ptA = d[pt_idx]
    ptB = d[pt_idx+1]
    #print(pt_idx)
    #print(d)
    if(abs(ptA.x-ptB.x)*dbu  == wg_width):
      new_y = (ptA.y+ptB.y)/2
      d[pt_idx].y = new_y
      d[pt_idx+1].y = new_y
    elif(abs(ptA.y-ptB.y)*dbu  == wg_width):
      new_x = (ptA.x + ptB.x)/2
      d[pt_idx].x = new_x
      d[pt_idx+1].x = new_x
    #print(d)
  
    self.cell.shapes(LayerSiN).insert(Polygon(d))
    
    # Create the pins, as short paths:
    w = int(round(self.wg_width/dbu))
    r = int(round(self.eff_r/dbu))
    x = r
    y = r
    
    # Pin on the top side:
    p2 = [Point(0, y-pin_length/2), Point(0, y+pin_length/2)]
    p2c = Point(x, y)
    self.set_p2 = p2c
    self.p2 = p2c
    pin = Path(p2, w)
    self.cell.shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, 0, y)
    text = Text ("pin2", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    x = 0
    y = 0
    # Pin on the left side:
    p1 = [Point(pin_length/2-r,0), Point(-pin_length/2-r,0)]
    p1c = Point(x,0)
    self.set_p1 = p1c
    self.p1 = p1c
    pin = Path(p1, w)
    self.cell.shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, -r, 0)
    text = Text ("pin1", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    points = arc_bezier(r, 270, 360, bezier)
    self.cell.shapes(LayerDevRecN).insert(pya.Path(points, w*3))



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class Bragg_Straight_from_file(pya.PCellDeclarationHelper):
  """
  by Rui Cheng 2020, UBC; modified by Lukas Chrostowski
  Description: Load Bragg grating waveguide vertices from a TXT file and draw

  Input: file with the format: (x1,y1,x2,y2;...)
  which can be generated in Matlab by:
    co = co / dbu;
    fid=fopen(file,'w');
    fprintf(fid,'(');
    fprintf(fid,'%d,%d;',round(co));
    fprintf(fid,')');
    fclose(fid);

  """

  def __init__(self):

    # Important: initialize the super class
    super(Bragg_Straight_from_file, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("forder", self.TypeString, "Forder", default = "")
    self.param("name", self.TypeString, "Name", default = "Square_BP_B3.5_N2536_DW6")  
    self.param("filetype", self.TypeString, "File type (txt)", default = 'txt')  
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("port_w", self.TypeDouble, "Port Waveguide width", default = 0.5)
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = TECHNOLOGY['Text'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
        return "Bragg_Straight_from_file_%s" % \
        (self.name)
  
  def coerce_parameters_impl(self):
    pass


  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
  
    import os
    from SiEPIC.extend import to_itype
    
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    TextLayerN = ly.layer(self.textl)
    LayerPinRec  = ly.layer(self.pinrec)
    
    file_path = os.path.join(self.forder, self.name + '.' + self.filetype)   
    # print('Bragg_Straight_from_file: input file: ' + file_path)

    if not(os.path.isfile(file_path)):
      error = 'File not found: ' + file_path
      text = Text (error, Trans(0,0))
      shape = shapes(TextLayerN).insert(text)
      shape.text_size = 1/dbu
      box=Box(0,0, len(error)*shape.text_size, shape.text_size)
      shapes(LayerDevRecN).insert(box)
      return
    
    if 'txt' in self.filetype:
      # Format: (x1,y1;x2,y2;...), where x and y values are integers
      text_file = open(file_path, "r")
      lines = text_file.readlines()
    else:
      error = 'File type not available: ' + self.filetype
      text = Text (error, Trans(0,0))
      shape = shapes(TextLayerN).insert(text)
      shape.text_size = 1/dbu
      box=Box(0,0, len(error)*shape.text_size, shape.text_size)
      shapes(LayerDevRecN).insert(box)
      return
        
    p = SimplePolygon().from_s(lines[0])
    shapes(LayerSiN).insert(p)
    
    # Create the pins on the waveguides, as short paths:
    from SiEPIC._globals import PIN_LENGTH as pin_length
    
    bbox = p.bbox()
    y = (bbox.p1.y + bbox.p2.y) / 2
    leftx = bbox.left
    port_w = to_itype(self.port_w,dbu)
    pin = Path([Point(pin_length/2, 0), Point(-pin_length/2, 0)], port_w)

    t = Trans(Trans.R0, bbox.left, y)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    t = Trans(Trans.R180, bbox.right, y)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    
    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    box=Box(bbox.left,y-3/2*port_w,bbox.right,y+3/2*port_w)
    shapes(LayerDevRecN).insert(box)

